""" Test the CLI for mip-rna link"""

import logging

from cg.cli.workflow.mip_rna.base import link


def test_cg_workflow_mip_rna_link(cli_runner, caplog, case_id, mip_rna_context):
    """Test print the MIP rna case config command to console"""

    caplog.set_level(logging.INFO)

    # GIVEN a store with a RNA case with wts application

    # WHEN we run a case in dry run mode
    result = cli_runner.invoke(link, [case_id], obj=mip_rna_context)

    # THEN the command should be printed
    assert result.exit_code == 0



from cg.exc import CgError


class PostProcessingRunFileManagerError(CgError):
    pass


class PostProcessingRunDataGeneratorError(CgError):
    pass


class PostProcessingParsingError(CgError):
    pass


class PostProcessingDataTransferError(CgError):
    pass


class PostProcessingStoreDataError(CgError):
    pass


class PostProcessingStoreFileError(CgError):
    pass


class PostProcessingError(CgError):
    pass



"""Tests for the PacBioPostprocessingService."""

from unittest import mock
from unittest.mock import Mock

import pytest

from cg.models.cg_config import CGConfig
from cg.services.run_devices.exc import (
    PostProcessingError,
    PostProcessingStoreDataError,
    PostProcessingStoreFileError,
)
from cg.services.run_devices.pacbio.data_storage_service.pacbio_store_service import (
    PacBioStoreService,
)
from cg.services.run_devices.pacbio.housekeeper_service.pacbio_houskeeper_service import (
    PacBioHousekeeperService,
)
from cg.services.run_devices.pacbio.post_processing_service import PacBioPostProcessingService


def test_pac_bio_post_processing_run_name_error(pac_bio_context):
    # GIVEN a PacBioPostProcessingService and a wrong run name
    run_name: str = "run_name"
    post_processing_service: PacBioPostProcessingService = (
        pac_bio_context.post_processing_services.pacbio
    )

    # WHEN storing post-processing data

    # THEN a PostProcessingError is raised
    with pytest.raises(PostProcessingError):
        post_processing_service.post_process(run_name=run_name)


def test_pac_bio_post_processing_store_data_error(
    pac_bio_context: CGConfig, pac_bio_sequencing_run_name: str
):
    # GIVEN a PacBioPostProcessingService and a wrong run name

    post_processing_service: PacBioPostProcessingService = (
        pac_bio_context.post_processing_services.pacbio
    )

    # WHEN storing post-processing data raises an error
    with mock.patch.object(
        PacBioStoreService, "store_post_processing_data", side_effect=PostProcessingStoreDataError
    ):
        # THEN a PostProcessingError is raised
        with pytest.raises(PostProcessingError):
            post_processing_service.post_process(run_name=pac_bio_sequencing_run_name)


def test_pac_bio_post_processing_store_files_error(
    pac_bio_context: CGConfig, pac_bio_sequencing_run_name: str
):
    # GIVEN a PacBioPostProcessingService
    post_processing_service: PacBioPostProcessingService = (
        pac_bio_context.post_processing_services.pacbio
    )
    post_processing_service.store_service = Mock()

    # WHEN storing post-processing files raises an error
    with mock.patch.object(
        PacBioHousekeeperService,
        "store_files_in_housekeeper",
        side_effect=PostProcessingStoreFileError,
    ):
        # THEN a PostProcessingError is raised
        with pytest.raises(PostProcessingError):
            post_processing_service.post_process(run_name=pac_bio_sequencing_run_name)



import logging
import os
import shutil
from datetime import datetime
from pathlib import Path
from subprocess import CalledProcessError
from typing import Iterator

import click
from housekeeper.store.models import Bundle, Version

from cg.apps.environ import environ_email
from cg.clients.chanjo2.models import CoverageMetrics
from cg.constants import EXIT_FAIL, EXIT_SUCCESS, Priority, SequencingFileTag, Workflow
from cg.constants.constants import (
    AnalysisType,
    CaseActions,
    FileFormat,
    GenomeVersion,
    WorkflowManager,
)
from cg.constants.gene_panel import GenePanelCombo, GenePanelMasterList
from cg.constants.scout import HGNC_ID, ScoutExportFileName
from cg.constants.tb import AnalysisStatus
from cg.exc import AnalysisNotReadyError, BundleAlreadyAddedError, CgDataError, CgError
from cg.io.controller import WriteFile
from cg.meta.archive.archive import SpringArchiveAPI
from cg.meta.meta import MetaAPI
from cg.meta.workflow.fastq import FastqHandler
from cg.models.analysis import AnalysisModel
from cg.models.cg_config import CGConfig
from cg.models.fastq import FastqFileMeta
from cg.services.sequencing_qc_service import SequencingQCService
from cg.store.models import Analysis, BedVersion, Case, CaseSample, Sample

LOG = logging.getLogger(__name__)


def add_gene_panel_combo(gene_panels: set[str]) -> set[str]:
    """Add gene panels combinations for gene panels being part of gene panel combination and return updated gene panels."""
    additional_panels = set()
    for panel in gene_panels:
        if panel in GenePanelCombo.COMBO_1:
            additional_panels |= GenePanelCombo.COMBO_1.get(panel)
    gene_panels |= additional_panels
    return gene_panels


class AnalysisAPI(MetaAPI):
    """
    Parent class containing all methods that are either shared or overridden by other workflow APIs.
    """

    def __init__(self, workflow: Workflow, config: CGConfig):
        super().__init__(config=config)
        self.workflow = workflow
        self._process = None

    @property
    def root(self):
        raise NotImplementedError

    @property
    def process(self):
        raise NotImplementedError

    @property
    def fastq_handler(self):
        return FastqHandler

    @property
    def is_multiple_samples_allowed(self) -> bool:
        """Return whether the analysis supports multiple samples to be linked to the case."""
        return True

    @staticmethod
    def get_help(context):
        """
        If no argument is passed, print help text
        """
        if context.invoked_subcommand is None:
            click.echo(context.get_help())

    def verify_deliverables_file_exists(self, case_id: str) -> None:
        if not Path(self.get_deliverables_file_path(case_id=case_id)).exists():
            raise CgError(f"No deliverables file found for case {case_id}")

    def verify_case_config_file_exists(self, case_id: str, dry_run: bool = False) -> None:
        if not Path(self.get_case_config_path(case_id=case_id)).exists() and not dry_run:
            raise CgError(f"No config file found for case {case_id}")

    def check_analysis_ongoing(self, case_id: str) -> None:
        if self.trailblazer_api.is_latest_analysis_ongoing(case_id=case_id):
            LOG.warning(f"{case_id} : analysis is still ongoing - skipping")
            raise CgError(f"Analysis still ongoing in Trailblazer for case {case_id}")

    def verify_case_path_exists(self, case_id: str) -> None:
        """Check if case path exists."""
        if not self.get_case_path(case_id=case_id).exists():
            LOG.info(f"No working directory for {case_id} exists")
            raise FileNotFoundError(f"No working directory for {case_id} exists")

    def is_case_ready_for_analysis(self, case: Case) -> bool:
        """Check if case is ready for analysis. If case passes sequencing QC and is set to analyze,
        or has not been analyzed yet, or the latest analysis failed, the case is ready for analysis.
        """
        case_passed_sequencing_qc: bool = SequencingQCService.case_pass_sequencing_qc(case)
        case_is_set_to_analyze: bool = case.action == CaseActions.ANALYZE
        case_has_not_been_analyzed: bool = not case.latest_analyzed
        case_latest_analysis_failed: bool = (
            self.trailblazer_api.get_latest_analysis_status(case_id=case.internal_id)
            == AnalysisStatus.FAILED
        )
        return case_passed_sequencing_qc and (
            case_is_set_to_analyze or case_has_not_been_analyzed or case_latest_analysis_failed
        )

    def get_cases_ready_for_analysis(self):
        """
        Return cases that are ready for analysis. The case is ready if it passes the logic in the
        get_cases_to_analyse method, and it has passed the pre-analysis quality check.
        """
        cases_to_analyse: list[Case] = self.get_cases_to_analyse()
        cases_passing_quality_check: list[Case] = [
            case for case in cases_to_analyse if SequencingQCService.case_pass_sequencing_qc(case)
        ]
        return cases_passing_quality_check

    def get_priority_for_case(self, case_id: str) -> int:
        """Get priority from the status db case priority"""
        case: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        return case.priority or Priority.research

    def get_slurm_qos_for_case(self, case_id: str) -> str:
        """Get Quality of service (SLURM QOS) for the case."""
        priority: int = self.get_priority_for_case(case_id=case_id)
        return Priority.priority_to_slurm_qos().get(priority)

    def get_workflow_manager(self) -> str:
        """Get workflow manager for a given workflow."""
        return WorkflowManager.Slurm.value

    def get_case_path(self, case_id: str) -> list[Path] | Path:
        """Path to case working directory."""
        raise NotImplementedError

    def get_case_config_path(self, case_id) -> Path:
        """Path to case config file"""
        raise NotImplementedError

    def get_job_ids_path(self, case_id: str) -> Path:
        """Path to file containing slurm/tower job ids for the case."""
        raise NotImplementedError

    def get_sample_name_from_lims_id(self, lims_id: str) -> str:
        """Retrieve sample name provided by customer for specific sample"""
        sample: Sample = self.status_db.get_sample_by_internal_id(internal_id=lims_id)
        return sample.name

    def link_fastq_files(self, case_id: str, dry_run: bool = False) -> None:
        """
        Links fastq files from Housekeeper to case working directory
        """
        raise NotImplementedError

    def get_bundle_deliverables_type(self, case_id: str) -> str | None:
        return None

    @staticmethod
    def get_application_type(sample_obj: Sample) -> str:
        """
        Gets application type for sample. Only application types supported by trailblazer (or other)
        are valid outputs.
        """
        prep_category: str = sample_obj.prep_category
        if prep_category and prep_category.lower() in {
            AnalysisType.TARGETED_GENOME_SEQUENCING,
            AnalysisType.WHOLE_EXOME_SEQUENCING,
            AnalysisType.WHOLE_GENOME_SEQUENCING,
            AnalysisType.WHOLE_TRANSCRIPTOME_SEQUENCING,
        }:
            return prep_category.lower()
        return AnalysisType.OTHER

    def get_case_application_type(self, case_id: str) -> str:
        """Returns the application type for samples in a case."""
        samples: list[Sample] = self.status_db.get_samples_by_case_id(case_id)
        application_types: set[str] = {self.get_application_type(sample) for sample in samples}

        if len(application_types) > 1:
            raise CgError(
                f"Different application_types found for case: {case_id} ({application_types})"
            )

        return application_types.pop()

    def are_case_samples_rna(self, case_id: str) -> bool:
        analysis_type: str = self.get_case_application_type(case_id)
        return analysis_type == AnalysisType.WHOLE_TRANSCRIPTOME_SEQUENCING

    def get_case_source_type(self, case_id: str) -> str | None:
        """
        Return the sample source type of a case.

        Raises:
            CgError: If different sources are set for the samples linked to a case.
        """
        sample_ids: Iterator[str] = self.status_db.get_sample_ids_by_case_id(case_id=case_id)
        source_types: set[str | None] = {
            self.lims_api.get_source(sample_id) for sample_id in sample_ids
        }
        if len(source_types) > 1:
            raise CgError(f"Different source types found for case: {case_id} ({source_types})")
        return source_types.pop()

    def has_case_only_exome_samples(self, case_id: str) -> bool:
        """Returns True if the application type for all samples in a case is WES."""
        application_type: str = self.get_case_application_type(case_id)
        return application_type == AnalysisType.WHOLE_EXOME_SEQUENCING

    def upload_bundle_housekeeper(
        self, case_id: str, dry_run: bool = False, force: bool = False
    ) -> None:
        """Storing bundle data in Housekeeper for a case."""
        LOG.info(f"Storing bundle data in Housekeeper for {case_id}")
        bundle_data: dict = self.get_hermes_transformed_deliverables(case_id=case_id, force=force)
        bundle_result: tuple[Bundle, Version] = self.housekeeper_api.add_bundle(
            bundle_data=bundle_data
        )
        if not bundle_result:
            LOG.info("Bundle already added to Housekeeper!")
            raise BundleAlreadyAddedError("Bundle already added to Housekeeper!")
        bundle_object, bundle_version = bundle_result
        if dry_run:
            LOG.info("Dry-run: Housekeeper changes will not be commited")
            LOG.info(
                "The following files would be stored:\n%s",
                "\n".join([f["path"] for f in bundle_data["files"]]),
            )
            return
        self.housekeeper_api.include(bundle_version)
        self.housekeeper_api.add_commit(bundle_object)
        self.housekeeper_api.add_commit(bundle_version)
        LOG.info(
            f"Analysis successfully stored in Housekeeper: {case_id} ({bundle_version.created_at})"
        )

    def upload_bundle_statusdb(
        self, case_id: str, comment: str | None = None, dry_run: bool = False, force: bool = False
    ) -> None:
        """Storing an analysis bundle in StatusDB for a provided case."""
        LOG.info(f"Storing analysis in StatusDB for {case_id}")
        case_obj: Case = self.status_db.get_case_by_internal_id(case_id)
        analysis_start: datetime.date = self.get_bundle_created_date(case_id)
        workflow_version: str = self.get_workflow_version(case_id)
        new_analysis: Analysis = self.status_db.add_analysis(
            workflow=self.workflow,
            version=workflow_version,
            completed_at=datetime.now() if not force else None,
            primary=(len(case_obj.analyses) == 0),
            started_at=analysis_start,
            comment=comment,
        )
        new_analysis.case = case_obj
        if dry_run:
            LOG.info("Dry-run: StatusDB changes will not be commited")
            return
        self.status_db.session.add(new_analysis)
        self.status_db.session.commit()
        LOG.info(f"Analysis successfully stored in StatusDB: {case_id} : {analysis_start}")

    def get_deliverables_file_path(self, case_id: str) -> Path:
        raise NotImplementedError

    def get_analysis_finish_path(self, case_id: str) -> Path:
        raise NotImplementedError

    def add_pending_trailblazer_analysis(self, case_id: str) -> None:
        self.check_analysis_ongoing(case_id)
        application_type: str = self.get_application_type(
            self.status_db.get_case_by_internal_id(case_id).links[0].sample
        )
        config_path: str = self.get_job_ids_path(case_id).as_posix()
        email: str = environ_email()
        order_id: int = self._get_order_id_from_case_id(case_id)
        out_dir: str = self.get_job_ids_path(case_id).parent.as_posix()
        slurm_quality_of_service: str = self.get_slurm_qos_for_case(case_id)
        ticket: str = self.status_db.get_latest_ticket_from_case(case_id)
        workflow: Workflow = self.workflow
        workflow_manager: str = self.get_workflow_manager()
        self.trailblazer_api.add_pending_analysis(
            analysis_type=application_type,
            case_id=case_id,
            config_path=config_path,
            email=email,
            order_id=order_id,
            out_dir=out_dir,
            slurm_quality_of_service=slurm_quality_of_service,
            ticket=ticket,
            workflow=workflow,
            workflow_manager=workflow_manager,
        )

    def _get_order_id_from_case_id(self, case_id) -> int:
        case: Case = self.status_db.get_case_by_internal_id(case_id)
        return case.latest_order.id

    def get_hermes_transformed_deliverables(self, case_id: str, force: bool = False) -> dict:
        """Return the Hermes-transformed deliverables bundle to be stored in Housekeeper for a given case."""
        return self.hermes_api.create_housekeeper_bundle(
            bundle_name=case_id,
            deliverables=self.get_deliverables_file_path(case_id=case_id),
            workflow=str(self.workflow),
            analysis_type=self.get_bundle_deliverables_type(case_id),
            created=self.get_bundle_created_date(case_id),
            force=force,
        ).model_dump()

    def get_bundle_created_date(self, case_id: str) -> datetime.date:
        """Return the creation date of the deliverables file for a given case."""
        return self.get_date_from_file_path(self.get_deliverables_file_path(case_id=case_id))

    def get_workflow_version(self, case_id: str) -> str:
        """
        Calls the workflow to get the workflow version number. If fails, returns a placeholder value instead.
        """
        try:
            self.process.run_command(["--version"])
            return list(self.process.stdout_lines())[0].split()[-1]
        except (Exception, CalledProcessError):
            LOG.warning(f"Could not retrieve {self.workflow} workflow version!")
            return "0.0.0"

    def set_statusdb_action(self, case_id: str, action: str | None, dry_run: bool = False) -> None:
        """
        Set one of the allowed actions on a case in StatusDB.
        """
        if dry_run:
            LOG.info(f"Dry-run: Action {action} would be set for case {case_id}")
            return
        if action in [None, *CaseActions.actions()]:
            case: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
            case.action = action
            self.status_db.session.commit()
            LOG.info("Action %s set for case %s", action, case_id)
            return
        LOG.warning(
            f"Action '{action}' not permitted by StatusDB and will not be set for case {case_id}"
        )

    def get_analyses_to_clean(self, before: datetime) -> list[Analysis]:
        analyses_to_clean = self.status_db.get_analyses_to_clean(
            before=before, workflow=self.workflow
        )
        return analyses_to_clean

    def get_cases_to_analyse(self) -> list[Case]:
        return self.status_db.cases_to_analyse(workflow=self.workflow)

    def get_cases_to_store(self) -> list[Case]:
        """Return cases where analysis finished successfully,
        and is ready to be stored in Housekeeper."""
        return [
            case
            for case in self.status_db.get_running_cases_in_workflow(workflow=self.workflow)
            if self.trailblazer_api.is_latest_analysis_completed(case_id=case.internal_id)
        ]

    def get_cases_to_qc(self) -> list[Case]:
        """Return cases where analysis finished successfully,
        and is ready for QC metrics checks."""
        return [
            case
            for case in self.status_db.get_running_cases_in_workflow(workflow=self.workflow)
            if self.trailblazer_api.is_latest_analysis_qc(case_id=case.internal_id)
        ]

    def get_sample_fastq_destination_dir(self, case: Case, sample: Sample) -> Path:
        """Return the path to the FASTQ destination directory."""
        raise NotImplementedError

    def gather_file_metadata_for_sample(self, sample: Sample) -> list[FastqFileMeta]:
        return [
            self.fastq_handler.parse_file_data(hk_file.full_path)
            for hk_file in self.housekeeper_api.files(
                bundle=sample.internal_id, tags={SequencingFileTag.FASTQ}
            )
        ]

    def get_validated_case(self, case_id: str) -> Case:
        case: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        if not case.links:
            raise CgError(f"No samples linked to {case_id}")
        if nlinks := len(case.links) > 1 and not self.is_multiple_samples_allowed:
            raise CgError(f"Only one sample per case is allowed. {nlinks} found")
        return case

    def link_fastq_files_for_sample(
        self, case: Case, sample: Sample, concatenate: bool = False
    ) -> None:
        """
        Link FASTQ files for a sample to the work directory.
        If workflow input requires concatenated fastq, files can also be concatenated
        """
        linked_reads_paths: dict[int, list[Path]] = {1: [], 2: []}
        concatenated_paths: dict[int, str] = {1: "", 2: ""}
        fastq_files_meta: list[FastqFileMeta] = self.gather_file_metadata_for_sample(sample=sample)
        sorted_fastq_files_meta: list[FastqFileMeta] = sorted(
            fastq_files_meta, key=lambda k: k.path
        )
        fastq_dir: Path = self.get_sample_fastq_destination_dir(case=case, sample=sample)
        fastq_dir.mkdir(parents=True, exist_ok=True)

        for fastq_file in sorted_fastq_files_meta:
            fastq_file_name: str = self.fastq_handler.create_fastq_name(
                lane=fastq_file.lane,
                flow_cell=fastq_file.flow_cell_id,
                sample=sample.internal_id,
                read_direction=fastq_file.read_direction,
                undetermined=fastq_file.undetermined,
                meta=self.get_lims_naming_metadata(sample),
            )
            destination_path = Path(fastq_dir, fastq_file_name)
            linked_reads_paths[fastq_file.read_direction].append(destination_path)
            concatenated_paths[fastq_file.read_direction] = (
                f"{fastq_dir}/{self.fastq_handler.get_concatenated_name(fastq_file_name)}"
            )

            if not destination_path.exists():
                LOG.info(f"Linking: {fastq_file.path} -> {destination_path}")
                destination_path.symlink_to(fastq_file.path)
            else:
                LOG.warning(f"Destination path already exists: {destination_path}")

        if not concatenate:
            return

        LOG.info(f"Concatenation in progress for sample: {sample.internal_id}")
        for read, value in linked_reads_paths.items():
            self.fastq_handler.concatenate(linked_reads_paths[read], concatenated_paths[read])
            self.fastq_handler.remove_files(value)

    def get_target_bed_from_lims(self, case_id: str) -> str | None:
        """Get target bed filename from LIMS."""
        case: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        sample: Sample = case.links[0].sample
        if sample.from_sample:
            sample: Sample = self.status_db.get_sample_by_internal_id(
                internal_id=sample.from_sample
            )
        target_bed_shortname: str | None = self.lims_api.capture_kit(lims_id=sample.internal_id)
        if not target_bed_shortname:
            return None
        bed_version: BedVersion | None = self.status_db.get_bed_version_by_short_name(
            bed_version_short_name=target_bed_shortname
        )
        if not bed_version:
            raise CgDataError(f"Bed-version {target_bed_shortname} does not exist")
        return bed_version.filename

    def decompression_running(self, case_id: str) -> None:
        """Check if decompression is running for a case"""
        is_decompression_running: bool = self.prepare_fastq_api.is_spring_decompression_running(
            case_id
        )
        if not is_decompression_running:
            LOG.warning(
                "Decompression can not be started for %s",
                case_id,
            )
            return
        LOG.info(
            "Decompression is running for %s, analysis will be started when decompression is done",
            case_id,
        )
        self.set_statusdb_action(case_id=case_id, action="analyze")

    def decompress_case(self, case_id: str, dry_run: bool) -> None:
        """Decompress all possible fastq files for all samples in a case"""
        LOG.info(
            "The analysis for %s could not start, decompression is needed",
            case_id,
        )
        decompression_possible: bool = (
            self.prepare_fastq_api.can_at_least_one_sample_be_decompressed(case_id)
        )
        if not decompression_possible:
            self.decompression_running(case_id=case_id)
            return
        case_obj: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        link: CaseSample
        any_decompression_started = False
        for link in case_obj.links:
            sample_id: str = link.sample.internal_id
            if dry_run:
                LOG.info(
                    f"This is a dry run, therefore decompression for {sample_id} won't be started"
                )
                continue
            decompression_started: bool = self.prepare_fastq_api.compress_api.decompress_spring(
                sample_id
            )
            if decompression_started:
                any_decompression_started = True

        if not any_decompression_started:
            LOG.warning("Decompression failed to start for %s", case_id)
            return
        if not dry_run:
            self.set_statusdb_action(case_id=case_id, action="analyze")
        LOG.info("Decompression started for %s", case_id)

    def resolve_decompression(self, case_id: str, dry_run: bool) -> None:
        """
        Decompresses a case if needed and adds new fastq files to Housekeeper.
        """
        if self.prepare_fastq_api.is_spring_decompression_needed(case_id):
            self.decompress_case(case_id=case_id, dry_run=dry_run)
            return

        if self.prepare_fastq_api.is_spring_decompression_running(case_id):
            self.set_statusdb_action(case_id=case_id, action=CaseActions.ANALYZE)
            return

        self.prepare_fastq_api.add_decompressed_fastq_files_to_housekeeper(case_id)

    @staticmethod
    def get_date_from_file_path(file_path: Path) -> datetime.date:
        """
        Get date from deliverables path using date created metadata.
        """
        return datetime.fromtimestamp(int(os.path.getctime(file_path)))

    def get_lims_naming_metadata(self, sample: Sample) -> str | None:
        return None

    def get_latest_metadata(self, case_id: str) -> AnalysisModel:
        """Return the latest analysis metadata of a specific case."""
        raise NotImplementedError

    def get_genome_build(self, case_id: str) -> str:
        """Return the build version of the reference genome."""
        raise NotImplementedError

    def get_variant_callers(self, case_id: str) -> list[str]:
        """Return list of variant-calling filters used during analysis."""
        return []

    def get_pons(self, case_id: str) -> list[str]:
        """Return list of panel of normals used for analysis."""
        return []

    def parse_analysis(
        self, config_raw: dict, qc_metrics_raw: dict, sample_info_raw: dict
    ) -> AnalysisModel:
        """Parses output analysis files."""
        raise NotImplementedError

    def clean_analyses(self, case_id: str) -> None:
        """Add a cleaned at date for all analyses related to a case."""
        analyses: list = self.status_db.get_case_by_internal_id(internal_id=case_id).analyses
        LOG.info(f"Adding a cleaned at date for case {case_id}")
        for analysis_obj in analyses:
            analysis_obj.cleaned_at = analysis_obj.cleaned_at or datetime.now()
            self.status_db.session.commit()

    def clean_run_dir(
        self, case_id: str, skip_confirmation: bool, case_path: list[Path] | Path
    ) -> int:
        """Remove workflow run directory."""

        try:
            self.verify_case_path_exists(case_id=case_id)
        except FileNotFoundError:
            self.clean_analyses(case_id)

        if skip_confirmation or click.confirm(
            f"Are you sure you want to remove all files in {case_path}?"
        ):
            if case_path.is_symlink():
                LOG.warning(
                    f"Will not automatically delete symlink: {case_path}, delete it manually",
                )
                return EXIT_FAIL

            shutil.rmtree(case_path, ignore_errors=True)
            LOG.info(f"Cleaned {case_path}")
            self.clean_analyses(case_id=case_id)
            return EXIT_SUCCESS

    def _is_illumina_run_check_applicable(self, case_id) -> bool:
        """Returns true if the case is neither down sampled nor external."""
        return not (
            self.status_db.is_case_down_sampled(case_id) or self.status_db.is_case_external(case_id)
        )

    def ensure_illumina_run_on_disk(self, case_id: str) -> None:
        """Check if Illumina sequencing runs are on disk for given case."""
        if not self._is_illumina_run_check_applicable(case_id):
            LOG.info(
                "Illumina run check is not applicable - "
                "the case is either down sampled or external."
            )
            return
        if not self.status_db.are_all_illumina_runs_on_disk(case_id=case_id):
            self.status_db.request_sequencing_runs_for_case(case_id)

    def is_raw_data_ready_for_analysis(self, case_id: str) -> bool:
        """Returns True if no files need to be retrieved from an external location and if all Spring files are
        decompressed."""
        if self.does_any_file_need_to_be_retrieved(case_id):
            return False
        if self.prepare_fastq_api.is_spring_decompression_needed(
            case_id
        ) or self.prepare_fastq_api.is_spring_decompression_running(case_id):
            LOG.warning(f"Case {case_id} is not ready - not all files are decompressed.")
            return False
        return True

    def does_any_file_need_to_be_retrieved(self, case_id: str) -> bool:
        """Checks whether we need to retrieve files from an external data location."""
        if self._is_illumina_run_check_applicable(
            case_id
        ) and not self.status_db.are_all_illumina_runs_on_disk(case_id):
            LOG.warning(f"Case {case_id} is not ready - not all Illumina runs present on disk.")
            return True
        else:
            if not self.are_all_spring_files_present(case_id):
                LOG.warning(f"Case {case_id} is not ready - some files are archived.")
                return True
        return False

    def prepare_fastq_files(self, case_id: str, dry_run: bool) -> None:
        """Retrieves or decompresses Spring files if needed. If so, an AnalysisNotReady error
        is raised."""
        self.ensure_files_are_present(case_id)
        self.resolve_decompression(case_id=case_id, dry_run=dry_run)
        if not self.is_raw_data_ready_for_analysis(case_id):
            raise AnalysisNotReadyError("FASTQ files are not present for the analysis to start")

    def ensure_files_are_present(self, case_id: str):
        """Checks if any Illumina runs need to be retrieved and submits a job if that is the case.
        Also checks if any spring files are archived and submits a job to retrieve any which are."""
        self.ensure_illumina_run_on_disk(case_id)
        if not self.are_all_spring_files_present(case_id):
            LOG.warning(f"Files are archived for case {case_id}")
            spring_archive_api = SpringArchiveAPI(
                status_db=self.status_db,
                housekeeper_api=self.housekeeper_api,
                data_flow_config=self.config.data_flow,
            )
            spring_archive_api.retrieve_spring_files_for_case(case_id)

    def are_all_spring_files_present(self, case_id: str) -> bool:
        """Return True if no Spring files for the case are archived in the data location used by the customer."""
        case: Case = self.status_db.get_case_by_internal_id(case_id)
        for sample in [link.sample for link in case.links]:
            if (
                files := self.housekeeper_api.get_archived_files_for_bundle(
                    bundle_name=sample.internal_id, tags=[SequencingFileTag.SPRING]
                )
            ) and not all(file.archive.retrieved_at for file in files):
                return False
        return True

    def get_archive_location_for_case(self, case_id: str) -> str:
        return self.status_db.get_case_by_internal_id(case_id).customer.data_archive_location

    @staticmethod
    def _write_managed_variants(out_dir: Path, content: list[str]) -> None:
        """Write the managed variants to case dir."""
        out_dir.mkdir(parents=True, exist_ok=True)
        WriteFile.write_file_from_content(
            content="\n".join(content),
            file_format=FileFormat.TXT,
            file_path=Path(out_dir, ScoutExportFileName.MANAGED_VARIANTS),
        )

    @staticmethod
    def _write_panel(out_dir: Path, content: list[str]) -> None:
        """Write the gene panel to case dir."""
        out_dir.mkdir(parents=True, exist_ok=True)
        WriteFile.write_file_from_content(
            content="\n".join(content),
            file_format=FileFormat.TXT,
            file_path=Path(out_dir, ScoutExportFileName.PANELS),
        )

    def _get_gene_panel(self, case_id: str, genome_build: str, dry_run: bool = False) -> list[str]:
        """Create and return the aggregated gene panel file."""
        case: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        all_panels: list[str] = self.get_aggregated_panels(
            customer_id=case.customer.internal_id, default_panels=set(case.panels)
        )
        return self.scout_api.export_panels(build=genome_build, panels=all_panels, dry_run=dry_run)

    def get_gene_panel(self, case_id: str, dry_run: bool = False) -> list[str]:
        """Create and return the aggregated gene panel file."""
        raise NotImplementedError

    def _get_managed_variants(self, genome_build: str) -> list[str]:
        """Create and return the managed variants."""
        return self.scout_api.export_managed_variants(genome_build=genome_build)

    def write_panel(self, case_id: str, content: list[str]) -> None:
        """Write the gene panel to case dir."""
        self._write_panel(out_dir=Path(self.root, case_id), content=content)

    @staticmethod
    def get_aggregated_panels(customer_id: str, default_panels: set[str]) -> list[str]:
        """Check if customer is collaborator for gene panel master list
        and if all default panels are included in the gene panel master list.
        If not, add gene panel combo and broad non-specific gene panels.
        Return an aggregated gene panel."""
        if GenePanelMasterList.is_customer_collaborator_and_panels_in_gene_panels_master_list(
            customer_id=customer_id, gene_panels=default_panels
        ):
            return GenePanelMasterList.get_panel_names()
        all_panels: set[str] = add_gene_panel_combo(gene_panels=default_panels)
        all_panels |= GenePanelMasterList.get_non_specific_gene_panels()
        return list(all_panels)

    def run_analysis(self, *args, **kwargs):
        raise NotImplementedError

    def get_data_analysis_type(self, case_id: str) -> str | None:
        """
        Return data analysis type carried out.
        Raises:
            ValueError: If the samples in a case have not the same analysis type.
        """
        case: Case = self.get_validated_case(case_id)
        analysis_types: set[str] = {
            link.sample.application_version.application.analysis_type for link in case.links
        }
        if len(analysis_types) > 1:
            raise ValueError(
                f"Case samples have different analysis types {', '.join(analysis_types)}"
            )
        return analysis_types.pop() if analysis_types else None

    @staticmethod
    def translate_genome_reference(genome_version: GenomeVersion) -> GenomeVersion:
        """Translates a genome reference assembly to its corresponding alternate name."""
        translation_map = {
            GenomeVersion.GRCh37: GenomeVersion.HG19,
            GenomeVersion.GRCh38: GenomeVersion.HG38,
            GenomeVersion.HG19: GenomeVersion.GRCh37,
            GenomeVersion.HG38: GenomeVersion.GRCh38,
        }
        return translation_map.get(genome_version, genome_version)

    def get_sample_coverage(
        self, case_id: str, sample_id: str, gene_ids: list[int]
    ) -> CoverageMetrics | None:
        """Return sample coverage data from Chanjo2."""
        raise NotImplementedError

    def get_scout_upload_case_tags(self):
        """Return workflow specific upload case tags."""
        raise NotImplementedError

    def get_gene_ids_from_scout(self, panels: list[str]) -> list[int]:
        """Return HGNC IDs of genes from specified panels using the Scout API."""
        gene_ids: list[int] = []
        for panel in panels:
            genes: list[dict] = self.scout_api.get_genes(panel)
            gene_ids.extend(gene.get(HGNC_ID) for gene in genes if gene.get(HGNC_ID) is not None)
        return gene_ids



"""Constants for delivery."""

from cg.constants.constants import Workflow
from cg.constants.housekeeper_tags import (
    AlignmentFileTag,
    AnalysisTag,
    HermesFileTag,
)

ONLY_ONE_CASE_PER_TICKET: list[Workflow] = [
    Workflow.FASTQ,
    Workflow.MICROSALT,
    Workflow.MUTANT,
]

SKIP_MISSING: list[Workflow] = [
    Workflow.FASTQ,
    Workflow.MICROSALT,
    Workflow.MUTANT,
]

BALSAMIC_ANALYSIS_CASE_TAGS: list[set[str]] = [
    {"delivery-report"},
    {"multiqc-html"},
    {"metrics"},
    {"cnv-report"},
    {"coverage"},
    {"germline"},
    {"vcf-sv"},
    {"vcf-sv-index"},
    {"vcf-sv-research"},
    {"vcf-sv-research-index"},
    {"vcf-sv-clinical"},
    {"vcf-sv-clinical-index"},
    {"vcf-snv"},
    {"vcf-snv-index"},
    {"vcf-snv-research"},
    {"vcf-snv-research-index"},
    {"vcf-snv-clinical"},
    {"vcf-snv-clinical-index"},
    {"vcf2cytosure"},
]

BALSAMIC_ANALYSIS_SAMPLE_TAGS: list[set[str]] = [
    {AlignmentFileTag.CRAM},
    {AlignmentFileTag.CRAM_INDEX},
]

BALSAMIC_QC_ANALYSIS_CASE_TAGS: list[set[str]] = [
    {"delivery-report"},
    {"multiqc-html"},
]

BALSAMIC_QC_ANALYSIS_SAMPLE_TAGS: list[set[str]] = [
    {"qc-cram"},
    {"qc-cram-index"},
]

BALSAMIC_UMI_ANALYSIS_CASE_TAGS: list[set[str]] = [
    {"vcf-umi-snv"},
    {"vcf-umi-snv-index"},
    {"vcf-umi-snv-research"},
    {"vcf-umi-snv-research-index"},
    {"vcf-umi-snv-clinical"},
    {"vcf-umi-snv-clinical-index"},
]

BALSAMIC_UMI_ANALYSIS_CASE_TAGS.extend(BALSAMIC_ANALYSIS_CASE_TAGS)

BALSAMIC_UMI_ANALYSIS_SAMPLE_TAGS: list[set[str]] = [
    {"umi-cram"},
    {"umi-cram-index"},
]

BALSAMIC_UMI_ANALYSIS_SAMPLE_TAGS.extend(BALSAMIC_ANALYSIS_SAMPLE_TAGS)

CLINICAL_DELIVERY_TAGS: list[set[str]] = [{HermesFileTag.CLINICAL_DELIVERY}]

FASTQ_ANALYSIS_CASE_TAGS: list[set[str]] = []

FASTQ_ANALYSIS_SAMPLE_TAGS: list[set[str]] = [
    {"fastq"},
]

MIP_DNA_ANALYSIS_CASE_TAGS: list[set[str]] = [
    {"delivery-report"},
    {"multiqc-html"},
    {"vcf-clinical-sv-bin"},
    {"vcf-clinical-sv-bin-index"},
    {"vcf-research-sv-bin"},
    {"vcf-research-sv-bin-index"},
    {"gbcf"},
    {"gbcf-index"},
    {"snv-gbcf"},
    {"snv-gbcf-index"},
    {"snv-bcf"},
    {"snv-bcf-index"},
    {"sv-bcf"},
    {"sv-bcf-index"},
    {"vcf-snv-clinical"},
    {"vcf-snv-clinical-index"},
    {"vcf-snv-research"},
    {"vcf-snv-research-index"},
    {"vcf-sv-clinical"},
    {"vcf-sv-clinical-index"},
    {"vcf-sv-research"},
    {"vcf-sv-research-index"},
]

MIP_DNA_ANALYSIS_SAMPLE_TAGS: list[set[str]] = [
    {AlignmentFileTag.BAM},
    {AlignmentFileTag.BAM_BAI},
    {AlignmentFileTag.CRAM},
    {AlignmentFileTag.CRAM_INDEX},
]

MIP_RNA_ANALYSIS_CASE_TAGS: list[set[str]] = [
    {"fusion", "clinical", "pdf"},
    {"fusion", "research", "pdf"},
    {"fusion", "vcf"},
    {"fusion", "vcf-index"},
    {"vcf-snv-clinical"},
    {"vcf-snv-clinical-index"},
    {"vcf-snv-research"},
    {"vcf-snv-research-index"},
    {"multiqc-html"},
]

MIP_RNA_ANALYSIS_SAMPLE_TAGS: list[set[str]] = [
    {AnalysisTag.FUSION, AnalysisTag.STARFUSION},
    {AnalysisTag.FUSION, AnalysisTag.ARRIBA},
    {AlignmentFileTag.CRAM},
    {AlignmentFileTag.CRAM_INDEX},
    {AnalysisTag.FUSION, "vcf"},
    {AnalysisTag.FUSION, "vcf-index"},
    {"salmon-quant"},
]

MICROSALT_ANALYSIS_CASE_TAGS = [{"qc-report"}, {"typing-report"}]

MICROSALT_ANALYSIS_SAMPLE_TAGS: list[set[str]] = []

MUTANT_ANALYSIS_CASE_TAGS: list[set[str]] = [
    {"pangolin"},
    {"ks-delivery"},
]

MUTANT_ANALYSIS_SAMPLE_TAGS: list[set[str]] = [
    {"fastq"},
    {"vcf", "vcf-report", "fohm-delivery"},
]


PIPELINE_ANALYSIS_TAG_MAP: dict[Workflow, dict] = {
    Workflow.BALSAMIC: {
        "case_tags": BALSAMIC_ANALYSIS_CASE_TAGS,
        "sample_tags": BALSAMIC_ANALYSIS_SAMPLE_TAGS,
    },
    Workflow.BALSAMIC_QC: {
        "case_tags": BALSAMIC_QC_ANALYSIS_CASE_TAGS,
        "sample_tags": BALSAMIC_QC_ANALYSIS_SAMPLE_TAGS,
    },
    Workflow.BALSAMIC_UMI: {
        "case_tags": BALSAMIC_UMI_ANALYSIS_CASE_TAGS,
        "sample_tags": BALSAMIC_UMI_ANALYSIS_SAMPLE_TAGS,
    },
    Workflow.MIP_DNA: {
        "case_tags": MIP_DNA_ANALYSIS_CASE_TAGS,
        "sample_tags": MIP_DNA_ANALYSIS_SAMPLE_TAGS,
    },
    Workflow.MIP_RNA: {
        "case_tags": MIP_RNA_ANALYSIS_CASE_TAGS,
        "sample_tags": MIP_RNA_ANALYSIS_SAMPLE_TAGS,
    },
    Workflow.MICROSALT: {
        "case_tags": MICROSALT_ANALYSIS_CASE_TAGS,
        "sample_tags": MICROSALT_ANALYSIS_SAMPLE_TAGS,
    },
    Workflow.FASTQ: {
        "case_tags": FASTQ_ANALYSIS_CASE_TAGS,
        "sample_tags": FASTQ_ANALYSIS_SAMPLE_TAGS,
    },
    Workflow.MUTANT: {
        "case_tags": MUTANT_ANALYSIS_CASE_TAGS,
        "sample_tags": MUTANT_ANALYSIS_SAMPLE_TAGS,
    },
    Workflow.RAREDISEASE: {
        "case_tags": CLINICAL_DELIVERY_TAGS,
        "sample_tags": CLINICAL_DELIVERY_TAGS,
    },
    Workflow.RNAFUSION: {
        "case_tags": CLINICAL_DELIVERY_TAGS,
        "sample_tags": CLINICAL_DELIVERY_TAGS,
    },
    Workflow.TAXPROFILER: {
        "case_tags": CLINICAL_DELIVERY_TAGS,
        "sample_tags": CLINICAL_DELIVERY_TAGS,
    },
    Workflow.TOMTE: {
        "case_tags": CLINICAL_DELIVERY_TAGS,
        "sample_tags": CLINICAL_DELIVERY_TAGS,
    },
}

PIPELINE_ANALYSIS_OPTIONS = PIPELINE_ANALYSIS_TAG_MAP.keys()

INBOX_NAME: str = "inbox"
OUTBOX_NAME: str = "outbox"



from cg.apps.tb.api import TrailblazerAPI
from cg.apps.tb.models import TrailblazerAnalysis


class AnalysisService:
    def __init__(self, analysis_client: TrailblazerAPI):
        self.analysis_client = analysis_client

    def add_upload_job(self, case_id: str, slurm_id: int):
        analysis: TrailblazerAnalysis = self.analysis_client.get_latest_completed_analysis(case_id)
        self.analysis_client.add_upload_job_to_analysis(slurm_id=slurm_id, analysis_id=analysis.id)



"""Coverage analysis API for human clinical sequencing data."""

import requests

from cg.clients.chanjo2.models import CoveragePostRequest, CoveragePostResponse
from cg.clients.chanjo2.utils import handle_client_errors


class Chanjo2APIClient:
    """
    Chanjo2 API to communicate with a d4tools software in order to return coverage
    over genomic intervals.
    """

    def __init__(self, base_url: str):
        self.base_url = base_url
        self.headers = {
            "Content-Type": "application/json",
            "accept": "application/json",
        }

    @handle_client_errors
    def get_coverage(self, coverage_post_request: CoveragePostRequest) -> CoveragePostResponse:
        """Send a POST request to the coverage endpoint to retrieve gene coverage summary data."""
        endpoint: str = f"{self.base_url}/coverage/d4/genes/summary"
        post_data: dict = coverage_post_request.model_dump()
        response: requests.Response = requests.post(
            url=endpoint, headers=self.headers, json=post_data
        )
        response.raise_for_status()
        response_json: dict = response.json()
        return CoveragePostResponse.model_validate(response_json)



"""Test for the string utilities."""

import pytest

from cg.exc import CgError
from cg.utils.string import get_element_from_split


def test_get_element_from_split():

    # GIVEN a string with a seperator
    separated_string: str = "zero_one_two_three"

    # WHEN getting an element divided by a separator based on the position
    element: str = get_element_from_split(value=separated_string, element_position=2, split="_")

    # THEN the expected element is returned
    assert element == "two"


def test_get_element_from_split_error():

    # GIVEN a string with a seperator
    separated_string: str = "zero_one_two_three"

    # WHEN getting an element divided by a separator based on the position that is out of bounds
    with pytest.raises(CgError):
        get_element_from_split(value=separated_string, element_position=12, split="_")

        # THEN an error is raised



"""Filters for the PacBio SMRT cell model."""

from enum import Enum

from sqlalchemy.orm import Query


def filter_pac_bio_smrt_cell_by_internal_id(smrt_cells: Query, internal_id: str, **kwargs) -> Query:
    """Get a PacBio SMRT cell by its internal id."""
    return smrt_cells.filter_by(internal_id=internal_id)


def apply_pac_bio_smrt_cell_filters(
    smrt_cells: Query, filter_functions: list[callable], internal_id: str
) -> Query:
    """Apply filtering functions and return filtered results."""
    for function in filter_functions:
        smrt_cells: Query = function(
            smrt_cells=smrt_cells,
            internal_id=internal_id,
        )
    return smrt_cells


class PacBioSMRTCellFilter(Enum):
    """Define SMRT cell filter functions."""

    BY_INTERNAL_ID: callable = filter_pac_bio_smrt_cell_by_internal_id



from pydantic_settings import BaseSettings


class AppConfig(BaseSettings):
    gunicorn_workers: int = 4
    gunicorn_threads: int = 4
    gunicorn_bind: str
    gunicorn_timeout: int = 400
    cg_sql_database_uri: str
    cg_secret_key: str = "thisIsNotASafeKey"
    invoice_max_price: int = 750_000
    lims_host: str
    lims_username: str
    lims_password: str
    osticket_api_key: str
    osticket_domain: str
    support_system_email: str
    email_uri: str
    google_oauth_client_id: str
    google_oauth_client_secret: str
    trailblazer_host: str
    trailblazer_service_account: str
    trailblazer_service_account_auth_file: str


app_config = AppConfig()



import datetime as dt

import pytest

from cg.constants import Workflow
from cg.constants.constants import PrepCategory
from cg.constants.subject import PhenotypeStatus
from cg.store.models import CaseSample
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


@pytest.fixture(name="store_with_analyses_for_cases")
def store_with_analyses_for_cases(
    analysis_store: Store,
    helpers: StoreHelpers,
    timestamp_now: dt.datetime,
    timestamp_yesterday: dt.datetime,
) -> Store:
    """Return a store with two analyses for two cases."""
    case_one = analysis_store.get_case_by_internal_id("yellowhog")
    case_two = helpers.add_case(analysis_store, internal_id="test_case_1")

    cases = [case_one, case_two]
    for case in cases:
        oldest_analysis = helpers.add_analysis(
            analysis_store,
            case=case,
            started_at=timestamp_yesterday,
            completed_at=timestamp_yesterday,
            uploaded_at=timestamp_yesterday,
            delivery_reported_at=None,
        )
        helpers.add_analysis(
            analysis_store,
            case=case,
            started_at=timestamp_now,
            completed_at=timestamp_now,
            uploaded_at=timestamp_now,
            delivery_reported_at=None,
        )
        sample = helpers.add_sample(analysis_store, delivered_at=timestamp_now)
        link: CaseSample = analysis_store.relate_sample(
            case=oldest_analysis.case, sample=sample, status=PhenotypeStatus.UNKNOWN
        )
        analysis_store.session.add(link)

    return analysis_store



"""Tests for the illumina demux version service."""

from pathlib import Path

from cg.services.illumina.file_parsing.demux_version_service import (
    IlluminaDemuxVersionService,
)


def test_get_demux_software_version(
    demux_version_file: Path,
    illumina_demux_version_service: IlluminaDemuxVersionService,
    expected_demux_software_version: str,
):
    """Test to get the demultiplexing software version."""
    # GIVEN a path to a demultiplexing log file

    # WHEN getting the demultiplexing software version
    demux_software_version: str = illumina_demux_version_service.get_demux_software_version(
        demux_version_file
    )

    # THEN assert that the demultiplexing software version is correct
    assert demux_software_version == expected_demux_software_version


def test_get_demux_software(
    demux_version_file: Path,
    illumina_demux_version_service: IlluminaDemuxVersionService,
    expected_demux_software: str,
):
    """Test to get the demultiplexing software."""
    # GIVEN a path to a demultiplexing log file

    # WHEN getting the demultiplexing software
    demux_software: str = illumina_demux_version_service.get_demux_software(demux_version_file)

    # THEN assert that the demultiplexing software is correct
    assert demux_software == expected_demux_software



from pathlib import Path

import pytest

from cg.io.config import write_config_nextflow_style


def test_write_config_nextflow_style(config_dict: Path):
    """
    Test output content from stream into nextflow config format.
    """

    # THEN assert a config format is returned
    assert (
        write_config_nextflow_style(content=config_dict)
        == 'params.input = "input_path"\nparams.output = "output_path"\n'
    )



import logging

from sqlalchemy.orm import Session

from cg.store.crud.create import CreateHandler
from cg.store.crud.delete import DeleteDataHandler
from cg.store.crud.read import ReadHandler
from cg.store.crud.update import UpdateHandler
from cg.store.database import get_session

LOG = logging.getLogger(__name__)


class Store(
    CreateHandler,
    DeleteDataHandler,
    ReadHandler,
    UpdateHandler,
):
    def __init__(self):
        self.session: Session = get_session()
        DeleteDataHandler(self.session)
        ReadHandler(self.session)
        UpdateHandler(self.session)

    def commit_to_store(self):
        """Commit pending changes to the store."""
        self.session.commit()

    def rollback(self):
        """Rollback any pending change to the store."""
        self.session.rollback()



"""RAREDISEASE upload API."""

import datetime as dt
import logging

import click

from cg.cli.generate.report.base import generate_delivery_report
from cg.cli.upload.clinical_delivery import upload_clinical_delivery
from cg.cli.upload.genotype import upload_genotypes
from cg.cli.upload.gens import upload_to_gens
from cg.cli.upload.observations import upload_observations_to_loqusdb
from cg.cli.upload.scout import upload_to_scout
from cg.constants import REPORT_SUPPORTED_DATA_DELIVERY, DataDelivery
from cg.meta.upload.upload_api import UploadAPI
from cg.meta.workflow.raredisease import RarediseaseAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.store.models import Analysis, Case

LOG = logging.getLogger(__name__)


class RarediseaseUploadAPI(UploadAPI):
    """RAREDISEASE upload API."""

    def __init__(self, config: CGConfig):
        self.analysis_api = RarediseaseAnalysisAPI(config)
        super().__init__(config=config, analysis_api=self.analysis_api)

    def upload(self, ctx: click.Context, case: Case, restart: bool) -> None:
        """Uploads RAREDISEASE analysis data and files."""
        analysis: Analysis = case.analyses[0]
        self.update_upload_started_at(analysis=analysis)

        ctx.invoke(upload_observations_to_loqusdb, case_id=case.internal_id)
        ctx.invoke(upload_to_gens, case_id=case.internal_id)

        # Clinical delivery upload
        ctx.invoke(upload_clinical_delivery, case_id=case.internal_id)

        LOG.info(
            f"Upload of case {case.internal_id} was successful. Uploaded at {dt.datetime.now()} in StatusDB"
        )
        self.update_uploaded_at(analysis=analysis)



from datetime import datetime

from cg.constants import SequencingRunDataAvailability
from cg.store.models import IlluminaSequencingRun, Sample
from cg.store.store import Store


def test_update_illumina_sequencing_run_availability(store_with_illumina_sequencing_data: Store):
    # GIVEN a store with Illumina Sequencing Runs that have data availability ON_DISK
    sequencing_run: IlluminaSequencingRun = (
        store_with_illumina_sequencing_data.get_illumina_sequencing_runs_by_data_availability(
            [SequencingRunDataAvailability.ON_DISK]
        )[0]
    )
    assert sequencing_run.data_availability == SequencingRunDataAvailability.ON_DISK

    # WHEN updating the data availability of a sequencing run
    store_with_illumina_sequencing_data.update_illumina_sequencing_run_data_availability(
        sequencing_run=sequencing_run, data_availability=SequencingRunDataAvailability.REQUESTED
    )

    # THEN the data availability of the sequencing run is updated
    assert sequencing_run.data_availability == SequencingRunDataAvailability.REQUESTED


def test_update_illumina_sequencing_run_has_backup(store_with_illumina_sequencing_data: Store):
    # GIVEN a store with Illumina Sequencing Runs that have backup status False
    sequencing_run: IlluminaSequencingRun = store_with_illumina_sequencing_data._get_query(
        IlluminaSequencingRun
    ).all()[0]
    assert sequencing_run.has_backup is False

    # WHEN updating the backup status of a sequencing run
    store_with_illumina_sequencing_data.update_illumina_sequencing_run_has_backup(
        sequencing_run=sequencing_run, has_backup=True
    )

    # THEN the backup status of the sequencing run is updated
    assert sequencing_run.has_backup is True


def test_update_sample_reads_illumina(
    store_with_illumina_sequencing_data: Store, selected_novaseq_x_sample_ids: list[str]
):
    # GIVEN a store with Illumina Sequencing Runs and a sample id
    sample: Sample = store_with_illumina_sequencing_data.get_sample_by_internal_id(
        selected_novaseq_x_sample_ids[0]
    )
    assert sample.reads == 0

    # WHEN updating the sample reads for a sequencing run
    store_with_illumina_sequencing_data.update_sample_reads_illumina(
        internal_id=selected_novaseq_x_sample_ids[0]
    )

    # THEN the total reads for the sample is updated
    total_reads_for_sample: int = 0
    sample_metrics: list[IlluminaSampleSequencingMetrics] = sample.sample_run_metrics
    for sample_metric in sample_metrics:
        total_reads_for_sample += sample_metric.total_reads_in_lane
    assert sample.reads == total_reads_for_sample


def test_update_sample_last_sequenced_at(
    store_with_illumina_sequencing_data: Store,
    selected_novaseq_x_sample_ids: list[str],
    timestamp_now: datetime,
):
    # GIVEN a store with Illumina Sequencing Runs and a sample id
    sample: Sample = store_with_illumina_sequencing_data.get_sample_by_internal_id(
        selected_novaseq_x_sample_ids[0]
    )
    assert sample.last_sequenced_at is None

    # WHEN updating the last sequenced at date for a sample
    store_with_illumina_sequencing_data.update_sample_sequenced_at(
        internal_id=selected_novaseq_x_sample_ids[0], date=timestamp_now
    )

    # THEN the last sequenced at date for the sample is updated
    assert sample.last_sequenced_at == timestamp_now



from http import HTTPStatus
from flask import Blueprint, abort, g, jsonify, request

from cg.server.dto.samples.collaborator_samples_request import CollaboratorSamplesRequest
from cg.server.dto.samples.collaborator_samples_request import CollaboratorSamplesRequest
from cg.server.dto.samples.samples_response import SamplesResponse
from cg.server.endpoints.utils import before_request
from cg.server.ext import sample_service
from cg.store.models import Customer, Sample
from cg.server.ext import db

SAMPLES_BLUEPRINT = Blueprint("samples", __name__, url_prefix="/api/v1")
SAMPLES_BLUEPRINT.before_request(before_request)


@SAMPLES_BLUEPRINT.route("/samples/<sample_id>")
def get_sample(sample_id):
    sample: Sample | None = db.get_sample_by_internal_id(sample_id)
    if sample is None:
        return abort(HTTPStatus.NOT_FOUND)
    if not g.current_user.is_admin and (sample.customer not in g.current_user.customers):
        return abort(HTTPStatus.FORBIDDEN)
    return jsonify(**sample.to_dict(links=True, flowcells=True))


@SAMPLES_BLUEPRINT.route("/samples_in_collaboration")
def get_samples_in_collaboration():
    data = CollaboratorSamplesRequest.model_validate(request.values.to_dict())
    response: SamplesResponse = sample_service.get_collaborator_samples(data)
    return jsonify(response.model_dump()), HTTPStatus.OK


@SAMPLES_BLUEPRINT.route("/samples_in_collaboration/<sample_id>")
def parse_sample_in_collaboration(sample_id):
    """Return a single sample."""
    sample: Sample = db.get_sample_by_internal_id(sample_id)
    customer: Customer = db.get_customer_by_internal_id(
        customer_internal_id=request.args.get("customer")
    )
    if sample.customer not in customer.collaborators:
        return abort(HTTPStatus.FORBIDDEN)
    return jsonify(**sample.to_dict(links=True, flowcells=True))


@SAMPLES_BLUEPRINT.route("/samples")
def get_samples():
    """Return samples."""
    if request.args.get("status") and not g.current_user.is_admin:
        return abort(HTTPStatus.FORBIDDEN)
    if request.args.get("status") == "incoming":
        samples: list[Sample] = db.get_samples_to_receive()
    elif request.args.get("status") == "labprep":
        samples: list[Sample] = db.get_samples_to_prepare()
    elif request.args.get("status") == "sequencing":
        samples: list[Sample] = db.get_samples_to_sequence()
    else:
        customers: list[Customer] | None = (
            None if g.current_user.is_admin else g.current_user.customers
        )
        samples: list[Sample] = db.get_samples_by_customer_id_and_pattern(
            pattern=request.args.get("enquiry"), customers=customers
        )
    limit = int(request.args.get("limit", 50))
    parsed_samples: list[dict] = [sample.to_dict() for sample in samples[:limit]]
    return jsonify(samples=parsed_samples, total=len(samples))



from pathlib import Path

from cg.services.run_devices.abstract_classes import RunDataGenerator
from cg.services.run_devices.error_handler import (
    handle_post_processing_errors,
)
from cg.services.run_devices.exc import PostProcessingRunDataGeneratorError
from cg.services.run_devices.pacbio.run_data_generator.run_data import PacBioRunData
from cg.services.run_devices.validators import (
    validate_has_expected_parts,
    validate_name_pre_fix,
)
from cg.utils.string import get_element_from_split


class PacBioRunDataGenerator(RunDataGenerator):

    @staticmethod
    def _validate_run_name(run_name) -> None:
        validate_name_pre_fix(run_name, pre_fix="r")
        validate_has_expected_parts(run_name=run_name, expected_parts=2)

    @handle_post_processing_errors(
        to_except=(ValueError,), to_raise=PostProcessingRunDataGeneratorError
    )
    def get_run_data(self, run_name: str, sequencing_dir: str) -> PacBioRunData:
        """
        Get the run data for a PacBio SMRT cell run.
        run_name should include the PacBio run including plate well, e.g. 'r84202_20240522_133539/1_A01'
        """
        self._validate_run_name(run_name)
        full_path = Path(sequencing_dir, run_name)

        return PacBioRunData(
            full_path=full_path,
            sequencing_run_name=self._get_sequencing_run_name(run_name),
            well_name=self._get_well(run_name),
            plate=self._get_plate(run_name),
        )

    @staticmethod
    def _get_sequencing_run_name(run_name: str) -> str:
        return get_element_from_split(value=run_name, element_position=0, split="/")

    @staticmethod
    def _get_plate_well(run_name: str) -> str:
        return get_element_from_split(value=run_name, element_position=-1, split="/")

    def _get_plate(self, run_name: str) -> str:
        plate_well: str = self._get_plate_well(run_name)
        return get_element_from_split(value=plate_well, element_position=0, split="_")

    def _get_well(self, run_name: str) -> str:
        plate_well: str = self._get_plate_well(run_name)
        return get_element_from_split(value=plate_well, element_position=-1, split="_")



"""Test module for the Rnafusion delivery report API."""

import pytest
from pytest_mock import MockFixture

from cg.constants import NA_FIELD, RIN_MAX_THRESHOLD, RIN_MIN_THRESHOLD
from cg.meta.report.rnafusion import RnafusionReportAPI
from cg.models.analysis import NextflowAnalysis
from cg.models.report.metadata import RnafusionSampleMetadataModel
from cg.store.models import Case, Sample
from tests.mocks.limsmock import MockLimsAPI


def test_get_sample_metadata(
    report_api_rnafusion: RnafusionReportAPI,
    sample_id: str,
    rnafusion_case_id: str,
    rnafusion_validated_metrics: dict[str, str],
    rnafusion_mock_analysis_finish: None,
):
    """Test Rnafusion sample metadata extraction."""

    # GIVEN a Rnafusion case and associated sample
    case: Case = report_api_rnafusion.status_db.get_case_by_internal_id(
        internal_id=rnafusion_case_id
    )
    sample: Sample = report_api_rnafusion.status_db.get_sample_by_internal_id(internal_id=sample_id)

    # GIVEN an analysis metadata object
    latest_metadata: NextflowAnalysis = report_api_rnafusion.analysis_api.get_latest_metadata(
        case_id=rnafusion_case_id
    )

    # WHEN getting the sample metadata
    sample_metadata: RnafusionSampleMetadataModel = report_api_rnafusion.get_sample_metadata(
        case=case, sample=sample, analysis_metadata=latest_metadata
    )

    # THEN the sample metadata should be correctly retrieved and match the expected validated metrics
    assert sample_metadata.model_dump() == rnafusion_validated_metrics


@pytest.mark.parametrize(
    "input_rin, expected_rin",
    [
        (RIN_MAX_THRESHOLD, str(float(RIN_MAX_THRESHOLD))),  # Test for a valid integer input
        (RIN_MAX_THRESHOLD + 1, NA_FIELD),  # Test for an integer above the allowed threshold
        (RIN_MIN_THRESHOLD - 1, NA_FIELD),  # Test for an integer below the allowed threshold
        (None, NA_FIELD),  # Test for a None input
    ],
)
def test_ensure_rin_thresholds(
    rnafusion_case_id: str,
    sample_id: str,
    input_rin: int | float,
    expected_rin: str,
    report_api_rnafusion: RnafusionReportAPI,
    rnafusion_mock_analysis_finish: None,
    mocker: MockFixture,
):
    """Test Rnafusion RIN value validation."""

    # GIVEN a Rnafusion case and associated sample
    case: Case = report_api_rnafusion.status_db.get_case_by_internal_id(
        internal_id=rnafusion_case_id
    )
    sample: Sample = report_api_rnafusion.status_db.get_sample_by_internal_id(internal_id=sample_id)

    # GIVEN an analysis metadata object
    latest_metadata: NextflowAnalysis = report_api_rnafusion.analysis_api.get_latest_metadata(
        case_id=rnafusion_case_id
    )

    # GIVEN a specific RIN value
    mocker.patch.object(MockLimsAPI, "get_sample_rin", return_value=input_rin)

    # WHEN getting the sample metadata
    sample_metadata: RnafusionSampleMetadataModel = report_api_rnafusion.get_sample_metadata(
        case=case, sample=sample, analysis_metadata=latest_metadata
    )

    # THEN the sample RIN value should match the expected RIN value
    assert sample_metadata.rin == expected_rin



from cg.services.orders.store_order_services.store_metagenome_order import (
    StoreMetagenomeOrderService,
)
from cg.services.orders.submitters.order_submitter import OrderSubmitter
from cg.services.orders.validate_order_services.validate_metagenome_order import (
    ValidateMetagenomeOrderService,
)


class MetagenomeOrderSubmitter(OrderSubmitter):
    """Class for submitting metagenome orders."""

    def __init__(
        self,
        order_validation_service: ValidateMetagenomeOrderService,
        order_store_service: StoreMetagenomeOrderService,
    ):
        self.order_validation_service = order_validation_service
        self.order_store_service = order_store_service

    def submit_order(self, order_in) -> dict:
        """Submit a metagenome order."""
        self.order_validation_service.validate_order(order_in)
        return self.order_store_service.store_order(order_in)



from cg.models.orders.excel_sample import ExcelSample


def are_all_samples_metagenome(samples: list[ExcelSample]) -> bool:
    """Check if all samples are metagenome samples"""
    return all(sample.application.startswith("ME") for sample in samples)



import pytest

from cg.cli.post_process.utils import get_post_processing_service_from_run_name
from cg.models.cg_config import CGConfig
from cg.services.run_devices.pacbio.post_processing_service import PacBioPostProcessingService


def test_get_post_processing_service_from_run_name(
    pac_bio_context: CGConfig, pac_bio_sequencing_run_name: str
):
    """Test that a the correct post processing service is returned given a run name."""
    # GIVEN a context with a post-processing service for the run name
    assert pac_bio_context.post_processing_services.pacbio

    # WHEN getting the post-processing service from the run name
    service: PacBioPostProcessingService = get_post_processing_service_from_run_name(
        context=pac_bio_context, run_name=pac_bio_sequencing_run_name
    )
    # THEN a the correct post-processing service should be returned
    assert isinstance(service, PacBioPostProcessingService)


@pytest.mark.parametrize(
    "wrong_run_name",
    [
        "m_84202_2024_05_22_133539/1_A01",
        "r84202_20240522133539/1_A01",
        "r84202_20240522_133539/3_A01",
    ],
    ids=["doesn't start with r", "incorrect splits", "incorrect plate"],
)
def test_get_post_processing_service_from_wrong_run_name(
    pac_bio_context: CGConfig, wrong_run_name: str
):
    """Test that a the correct post processing service is returned given a run name."""
    # GIVEN a wrong run name

    # GIVEN a context with a post-processing service for the run name
    assert pac_bio_context.post_processing_services.pacbio

    # WHEN getting the post-processing service from the run name
    with pytest.raises(NameError):
        get_post_processing_service_from_run_name(context=pac_bio_context, run_name=wrong_run_name)

    # THEN an error is raised



import logging
from http import HTTPStatus

from pydantic import ValidationError
from requests import RequestException, Response, Session
from requests.adapters import HTTPAdapter
from urllib3 import Retry

from cg.clients.freshdesk.constants import EndPoints
from cg.clients.freshdesk.exceptions import (
    FreshdeskAPIException,
    FreshdeskModelException,
)
from cg.clients.freshdesk.models import TicketCreate, TicketResponse

LOG = logging.getLogger(__name__)


class FreshdeskClient:
    """Client for communicating with the freshdesk REST API."""

    def __init__(self, base_url: str, api_key: str):
        self.base_url = base_url
        self.api_key = api_key
        self.session = self._get_session()

    def create_ticket(self, ticket: TicketCreate) -> TicketResponse:
        """Create a ticket."""
        LOG.debug(ticket.model_dump_json())
        try:
            response: Response = self.session.post(
                url=self._url(EndPoints.TICKETS),
                json=ticket.model_dump(exclude_none=True),
            )
            response.raise_for_status()
            return TicketResponse.model_validate(response.json())
        except RequestException as error:
            LOG.error(f"Could not create ticket: {error}")
            raise FreshdeskAPIException(error) from error
        except ValidationError as error:
            LOG.error(f"Response from Freshdesk does not fit model: {TicketResponse}.\n{error}")
            raise FreshdeskModelException(error) from error

    def _url(self, endpoint: str) -> str:
        """Get the full URL for the endpoint."""
        return f"{self.base_url}{endpoint}"

    def _get_session(self) -> Session:
        """Configures and sets a session to be used for requests."""
        session = Session()
        self._configure_retries(session)
        session.auth = (self.api_key, "X")
        session.headers.update({"Content-Type": "application/json"})
        return session

    @staticmethod
    def _configure_retries(session: Session) -> None:
        """Configures retries for the session."""
        retry_strategy = Retry(
            total=5,
            status_forcelist=[
                HTTPStatus.TOO_MANY_REQUESTS,
                HTTPStatus.INTERNAL_SERVER_ERROR,
                HTTPStatus.BAD_GATEWAY,
                HTTPStatus.SERVICE_UNAVAILABLE,
                HTTPStatus.GATEWAY_TIMEOUT,
            ],
            backoff_factor=2,
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        session.mount("https://", adapter)



"""Fixtures related to grouping all sequencing devices: PacBio, Illumina, ONT & Saphire."""

import pytest

from cg.models.cg_config import PostProcessingServices
from cg.services.run_devices.pacbio.post_processing_service import PacBioPostProcessingService


@pytest.fixture
def post_processing_services(
    pac_bio_post_processing_service: PacBioPostProcessingService,
) -> PostProcessingServices:
    """Return the post processing services."""
    return PostProcessingServices(
        pacbio=pac_bio_post_processing_service,
    )



from datetime import datetime


def parse_date(date_string: str) -> datetime:
    if len(date_string) == 6:
        return datetime.strptime(date_string, "%y%m%d")
    elif len(date_string) == 8:
        return datetime.strptime(date_string, "%Y%m%d")
    else:
        raise ValueError("Date format not recognized")



"""Module that holds the illumina post-processing service."""

import logging
from pathlib import Path

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import SequencingFileTag
from cg.constants.devices import DeviceType
from cg.exc import FlowCellError, MissingFilesError
from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData
from cg.services.illumina.data_transfer.data_transfer_service import (
    IlluminaDataTransferService,
)
from cg.services.illumina.data_transfer.models import (
    IlluminaFlowCellDTO,
    IlluminaSampleSequencingMetricsDTO,
    IlluminaSequencingRunDTO,
)
from cg.services.illumina.post_processing.housekeeper_storage import (
    add_demux_logs_to_housekeeper,
    add_run_parameters_file_to_housekeeper,
    add_sample_fastq_files_to_housekeeper,
    delete_sequencing_data_from_housekeeper,
    store_undetermined_fastq_files,
)
from cg.services.illumina.post_processing.utils import (
    combine_sample_metrics_with_undetermined,
    create_delivery_file_in_flow_cell_directory,
)
from cg.services.illumina.post_processing.validation import (
    is_flow_cell_ready_for_postprocessing,
)
from cg.store.exc import EntryNotFoundError
from cg.store.models import IlluminaFlowCell, IlluminaSequencingRun
from cg.store.store import Store
from cg.utils.files import get_directories_in_path

LOG = logging.getLogger(__name__)


class IlluminaPostProcessingService:
    def __init__(
        self,
        status_db: Store,
        housekeeper_api: HousekeeperAPI,
        demultiplexed_runs_dir: Path,
        dry_run: bool,
    ) -> None:
        self.status_db: Store = status_db
        self.hk_api: HousekeeperAPI = housekeeper_api
        self.demultiplexed_runs_dir = demultiplexed_runs_dir
        self.dry_run: bool = dry_run

    def store_illumina_flow_cell(
        self,
        run_directory_data: IlluminaRunDirectoryData,
    ) -> IlluminaFlowCell:
        """
        Create Illumina flow cell from the parsed and validated run directory data
        and add the run samples to the model.
        """
        model: str | None = run_directory_data.run_parameters.get_flow_cell_model()
        flow_cell_dto = IlluminaFlowCellDTO(
            internal_id=run_directory_data.id, type=DeviceType.ILLUMINA, model=model
        )
        return self.status_db.add_illumina_flow_cell(flow_cell_dto)

    def store_illumina_sequencing_run(
        self,
        run_directory_data: IlluminaRunDirectoryData,
        flow_cell: IlluminaFlowCell,
    ) -> IlluminaSequencingRun:
        """Store Illumina run metrics in the status database."""
        metrics_service = IlluminaDataTransferService()
        sequencing_run_dto: IlluminaSequencingRunDTO = (
            metrics_service.create_illumina_sequencing_dto(run_directory_data)
        )
        return self.status_db.add_illumina_sequencing_run(
            sequencing_run_dto=sequencing_run_dto, flow_cell=flow_cell
        )

    def store_illumina_sample_sequencing_metrics(
        self,
        run_directory_data: IlluminaRunDirectoryData,
        sequencing_run: IlluminaSequencingRun,
    ) -> list[IlluminaSampleSequencingMetricsDTO]:
        """Store Illumina sample sequencing metrics in the status database."""
        metrics_service = IlluminaDataTransferService()
        sample_metrics: list[IlluminaSampleSequencingMetricsDTO] = (
            metrics_service.create_sample_sequencing_metrics_dto_for_flow_cell(
                flow_cell_directory=run_directory_data.get_demultiplexed_runs_dir(),
            )
        )
        undetermined_metrics: list[IlluminaSampleSequencingMetricsDTO] = (
            metrics_service.create_sample_run_dto_for_undetermined_reads(run_directory_data)
        )
        combined_metrics: list[IlluminaSampleSequencingMetricsDTO] = (
            combine_sample_metrics_with_undetermined(
                sample_metrics=sample_metrics,
                undetermined_metrics=undetermined_metrics,
            )
        )
        for sample_metric in combined_metrics:
            self.status_db.add_illumina_sample_metrics_entry(
                metrics_dto=sample_metric, sequencing_run=sequencing_run
            )
        return combined_metrics

    def store_sequencing_data_in_status_db(
        self, run_directory_data: IlluminaRunDirectoryData
    ) -> None:
        """Store all Illumina sequencing data in the status database."""
        LOG.info(f"Add sequencing and demux data to StatusDB for run {run_directory_data.id}")
        flow_cell: IlluminaFlowCell = self.store_illumina_flow_cell(
            run_directory_data=run_directory_data
        )
        sequencing_run: IlluminaSequencingRun = self.store_illumina_sequencing_run(
            run_directory_data=run_directory_data, flow_cell=flow_cell
        )
        sample_metrics: list[IlluminaSampleSequencingMetricsDTO] = (
            self.store_illumina_sample_sequencing_metrics(
                run_directory_data=run_directory_data, sequencing_run=sequencing_run
            )
        )
        self.update_samples_for_metrics(
            sample_metrics=sample_metrics, sequencing_run=sequencing_run
        )
        self.status_db.commit_to_store()

    def update_samples_for_metrics(
        self,
        sample_metrics: list[IlluminaSampleSequencingMetricsDTO],
        sequencing_run: IlluminaSequencingRun,
    ) -> None:
        unique_samples_on_run: list[str] = self.get_unique_samples_from_run(sample_metrics)
        for sample_id in unique_samples_on_run:
            self.status_db.update_sample_reads_illumina(internal_id=sample_id)
            self.status_db.update_sample_sequenced_at(
                sample_id, date=sequencing_run.sequencing_completed_at
            )

    @staticmethod
    def get_unique_samples_from_run(
        sample_metrics: list[IlluminaSampleSequencingMetricsDTO],
    ) -> list[str]:
        """Get unique samples from the run."""
        return list({sample_metric.sample_id for sample_metric in sample_metrics})

    def store_sequencing_data_in_housekeeper(
        self,
        run_directory_data: IlluminaRunDirectoryData,
        store: Store,
    ) -> None:
        """Store fastq files, demux logs and run parameters for sequencing run in Housekeeper."""
        LOG.info(f"Add sequencing and demux data to Housekeeper for run {run_directory_data.id}")

        self.hk_api.add_bundle_and_version_if_non_existent(run_directory_data.id)
        tags: list[str] = [
            SequencingFileTag.FASTQ,
            SequencingFileTag.RUN_PARAMETERS,
            run_directory_data.id,
        ]
        self.hk_api.add_tags_if_non_existent(tags)
        add_sample_fastq_files_to_housekeeper(
            run_directory_data=run_directory_data, hk_api=self.hk_api, store=store
        )
        store_undetermined_fastq_files(
            run_directory_data=run_directory_data, hk_api=self.hk_api, store=store
        )
        add_demux_logs_to_housekeeper(
            run_directory_data=run_directory_data,
            hk_api=self.hk_api,
        )
        add_run_parameters_file_to_housekeeper(
            run_directory_data=run_directory_data,
            hk_api=self.hk_api,
        )

    def post_process_illumina_flow_cell(
        self,
        sequencing_run_name: str,
    ) -> None:
        """Store data for an Illumina demultiplexed run and mark it as ready for delivery.
        This function:
            - Stores the run data in the status database
            - Stores sequencing metrics in the status database
            - Updates sample read counts in the status database
            - Stores the run data in the Housekeeper database
            - Creates a delivery file in the sequencing run directory
        Raises:
            FlowCellError: If the flow cell directory or the data it contains is not valid.
        """

        LOG.info(f"Post-process Illumina run {sequencing_run_name}")
        demux_run_dir = Path(self.demultiplexed_runs_dir, sequencing_run_name)
        run_directory_data = IlluminaRunDirectoryData(demux_run_dir)
        sample_sheet_path: Path = self.hk_api.get_sample_sheet_path(run_directory_data.id)
        run_directory_data.set_sample_sheet_path_hk(hk_path=sample_sheet_path)
        sequencing_run: IlluminaSequencingRun | None = None
        has_backup: bool = False

        LOG.debug("Set path for Housekeeper sample sheet in run directory")
        try:
            is_flow_cell_ready_for_postprocessing(
                flow_cell_output_directory=demux_run_dir,
                flow_cell=run_directory_data,
            )
        except (FlowCellError, MissingFilesError) as e:
            LOG.warning(f"Run {sequencing_run_name} will be skipped: {e}")
            return
        if self.dry_run:
            LOG.info(f"Dry run: will not post-process Illumina run {sequencing_run_name}")
            return
        try:
            sequencing_run: IlluminaSequencingRun = (
                self.status_db.get_illumina_sequencing_run_by_device_internal_id(
                    run_directory_data.id
                )
            )
            has_backup: bool = sequencing_run.has_backup
        except EntryNotFoundError as error:
            LOG.info(f"Run {sequencing_run_name} not found in StatusDB: {str(error)}")
        self.delete_sequencing_run_data(flow_cell_id=run_directory_data.id)
        try:
            self.store_sequencing_data_in_status_db(run_directory_data)
            self.store_sequencing_data_in_housekeeper(
                run_directory_data=run_directory_data,
                store=self.status_db,
            )
        except Exception as e:
            LOG.error(f"Failed to store Illumina run: {str(e)}")
            raise
        if sequencing_run:
            self.status_db.update_illumina_sequencing_run_has_backup(
                sequencing_run=sequencing_run, has_backup=has_backup
            )

        create_delivery_file_in_flow_cell_directory(demux_run_dir)

    def get_all_demultiplexed_runs(self) -> list[Path]:
        """Get all demultiplexed Illumina runs."""
        return get_directories_in_path(self.demultiplexed_runs_dir)

    def post_process_all_runs(self) -> bool:
        """Post process all demultiplex illumina runs that need it."""
        demux_dirs = self.get_all_demultiplexed_runs()
        is_error_raised: bool = False
        for demux_dir in demux_dirs:
            try:
                self.post_process_illumina_flow_cell(demux_dir.name)
            except Exception as error:
                LOG.error(
                    f"Failed to post process demultiplexed Illumina run {demux_dir.name}: {str(error)}"
                )
                is_error_raised = True
                continue
        return is_error_raised

    def delete_sequencing_run_data(self, flow_cell_id: str):
        """Delete sequencing run entries from Housekeeper and StatusDB."""
        try:
            self.status_db.delete_illumina_flow_cell(flow_cell_id)
        except EntryNotFoundError:
            LOG.warning(f"Flow cell {flow_cell_id} not found in StatusDB.")
        delete_sequencing_data_from_housekeeper(flow_cell_id=flow_cell_id, hk_api=self.hk_api)



from pathlib import Path

import pytest

from cg.constants.housekeeper_tags import AlignmentFileTag
from cg.constants.pacbio import PacBioHousekeeperTags
from cg.services.run_devices.pacbio.housekeeper_service.models import PacBioFileData


@pytest.fixture
def ccs_report_pac_bio_file_data(
    pac_bio_ccs_report_file: Path, smrt_cell_internal_id: str
) -> PacBioFileData:
    """Return a PacBioFileData object for a CCS report file."""
    return PacBioFileData(
        file_path=pac_bio_ccs_report_file,
        bundle_name=smrt_cell_internal_id,
        tags=[PacBioHousekeeperTags.CCS_REPORT, smrt_cell_internal_id],
    )


@pytest.fixture
def pac_bio_hifi_reads_file_data(
    pac_bio_hifi_read_file: Path, smrt_cell_internal_id: str, pac_bio_sample_internal_id: str
) -> PacBioFileData:
    """Return a PacBioFileData object for a HiFi reads file."""
    return PacBioFileData(
        file_path=pac_bio_hifi_read_file,
        bundle_name=pac_bio_sample_internal_id,
        tags=[AlignmentFileTag.BAM, smrt_cell_internal_id, pac_bio_sample_internal_id],
    )



"""Helper functions for observations related actions."""

import logging

from sqlalchemy.orm import Query

from cg.constants.constants import Workflow
from cg.constants.observations import LOQUSDB_SUPPORTED_WORKFLOWS
from cg.exc import CaseNotFoundError
from cg.meta.observations.balsamic_observations_api import BalsamicObservationsAPI
from cg.meta.observations.mip_dna_observations_api import MipDNAObservationsAPI
from cg.meta.observations.raredisease_observations_api import RarediseaseObservationsAPI
from cg.models.cg_config import CGConfig
from cg.store.models import Case
from cg.store.store import Store

LOG = logging.getLogger(__name__)


def get_observations_verified_case(context: CGConfig, case_id: str | None, upload: bool) -> Case:
    """Return a verified Loqusdb case."""
    status_db: Store = context.status_db
    case: Case = status_db.get_case_by_internal_id(internal_id=case_id)
    if not case or case.data_analysis not in LOQUSDB_SUPPORTED_WORKFLOWS:
        LOG.error("Invalid case ID. Retrieving available cases for Loqusdb actions.")
        cases_to_process: Query = (
            status_db.observations_to_upload() if upload else status_db.observations_uploaded()
        )
        if not cases_to_process:
            LOG.info("There are no valid cases to be processed by Loqusdb")
        else:
            LOG.info("Provide one of the following case IDs: ")
            for case in cases_to_process:
                LOG.info(f"{case.internal_id} ({case.data_analysis})")
        raise CaseNotFoundError
    return case


def get_observations_api(
    context: CGConfig, case_id: str | None, upload: bool
) -> BalsamicObservationsAPI | MipDNAObservationsAPI | RarediseaseObservationsAPI:
    """Return an observations API given a specific case object."""
    case: Case = get_observations_verified_case(context=context, case_id=case_id, upload=upload)
    observations_apis = {
        Workflow.BALSAMIC: BalsamicObservationsAPI(context),
        Workflow.MIP_DNA: MipDNAObservationsAPI(context),
        Workflow.RAREDISEASE: RarediseaseObservationsAPI(context),
    }
    return observations_apis[case.data_analysis]



import logging
from pathlib import Path
from typing import Iterable

from housekeeper.store.models import File

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants.housekeeper_tags import SequencingFileTag
from cg.constants.sequencing import Sequencers
from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData
from cg.services.illumina.post_processing.utils import (
    get_lane_from_sample_fastq,
    get_q30_threshold,
    get_sample_fastqs_from_flow_cell,
    get_undetermined_fastqs,
    rename_fastq_file_if_needed,
)
from cg.store.store import Store
from cg.utils.files import get_files_matching_pattern

LOG = logging.getLogger(__name__)


def _should_fastq_path_be_stored_in_housekeeper(
    sample_id: str,
    sample_fastq_path: Path,
    sequencer_type: Sequencers,
    device_internal_id: str,
    store: Store,
) -> bool:
    """
    Check if a sample fastq file should be tracked in Housekeeper.
    Only fastq files that pass the q30 threshold should be tracked.
    """
    lane = get_lane_from_sample_fastq(sample_fastq_path)
    q30_threshold: int = get_q30_threshold(sequencer_type)

    metric = store.get_illumina_metrics_entry_by_device_sample_and_lane(
        device_internal_id=device_internal_id,
        sample_internal_id=sample_id,
        lane=lane,
    )

    if metric:
        return metric.base_passing_q30_percent >= q30_threshold

    LOG.warning(
        f"Skipping fastq file {sample_fastq_path.name} as no metrics entry was found in status db."
    )
    LOG.warning(f"Flow cell name: {device_internal_id}, sample id: {sample_id}, lane: {lane} ")
    return False


def add_sample_fastq_files_to_housekeeper(
    run_directory_data: IlluminaRunDirectoryData, hk_api: HousekeeperAPI, store: Store
) -> None:
    """Add sample fastq files from the demultiplex directory to Housekeeper."""
    sample_internal_ids: list[str] = run_directory_data.sample_sheet.get_sample_ids()
    for sample_internal_id in sample_internal_ids:
        sample_fastq_paths: list[Path] | None = get_sample_fastqs_from_flow_cell(
            demultiplexed_run_path=run_directory_data.get_demultiplexed_runs_dir(),
            sample_internal_id=sample_internal_id,
        )
        if not sample_fastq_paths:
            LOG.warning(
                f"Cannot find fastq files for sample {sample_internal_id} in "
                f"{run_directory_data.get_demultiplexed_runs_dir()}. Skipping."
            )
            continue
        for sample_fastq_path in sample_fastq_paths:
            sample_fastq_path: Path = rename_fastq_file_if_needed(
                fastq_file_path=sample_fastq_path, flow_cell_name=run_directory_data.id
            )
            if _should_fastq_path_be_stored_in_housekeeper(
                sample_id=sample_internal_id,
                sample_fastq_path=sample_fastq_path,
                sequencer_type=run_directory_data.sequencer_type,
                device_internal_id=run_directory_data.id,
                store=store,
            ):
                hk_api.create_bundle_and_add_file_with_tags(
                    bundle_name=sample_internal_id,
                    file_path=sample_fastq_path,
                    tags=[run_directory_data.id, sample_internal_id, SequencingFileTag.FASTQ],
                )


def store_undetermined_fastq_files(
    run_directory_data: IlluminaRunDirectoryData, hk_api: HousekeeperAPI, store: Store
) -> None:
    """Store undetermined fastq files for non-pooled samples in Housekeeper."""
    non_pooled_lanes_and_samples: list[tuple[int, str]] = (
        run_directory_data.sample_sheet.get_non_pooled_lanes_and_samples()
    )

    for lane, sample_id in non_pooled_lanes_and_samples:
        undetermined_fastqs: list[Path] = get_undetermined_fastqs(
            lane=lane, demultiplexed_run_path=run_directory_data.get_demultiplexed_runs_dir()
        )

        for fastq_path in undetermined_fastqs:
            if _should_fastq_path_be_stored_in_housekeeper(
                sample_id=sample_id,
                sample_fastq_path=fastq_path,
                sequencer_type=run_directory_data.sequencer_type,
                device_internal_id=run_directory_data.id,
                store=store,
            ):
                hk_api.create_bundle_and_add_file_with_tags(
                    bundle_name=sample_id,
                    file_path=fastq_path,
                    tags=[run_directory_data.id, SequencingFileTag.FASTQ, sample_id],
                )


def add_demux_logs_to_housekeeper(
    run_directory_data: IlluminaRunDirectoryData, hk_api: HousekeeperAPI
) -> None:
    """Add demux logs to Housekeeper."""
    log_file_name_pattern: str = r"*_demultiplex.std*"
    demux_log_file_paths: list[Path] = get_files_matching_pattern(
        directory=run_directory_data.get_sequencing_runs_dir(), pattern=log_file_name_pattern
    )

    tag_names: list[str] = [SequencingFileTag.DEMUX_LOG, run_directory_data.id]
    for log_file_path in demux_log_file_paths:
        try:
            hk_api.add_file_to_bundle_if_non_existent(
                file_path=log_file_path, bundle_name=run_directory_data.id, tag_names=tag_names
            )
            LOG.info(f"Added demux log file {log_file_path} to Housekeeper.")
        except FileNotFoundError as e:
            LOG.error(f"Cannot find demux log file {log_file_path}. Error: {e}.")


def add_run_parameters_file_to_housekeeper(
    run_directory_data: IlluminaRunDirectoryData, hk_api: HousekeeperAPI
) -> None:
    """Add run parameters file to Housekeeper."""
    run_parameters_file_path: Path = run_directory_data.run_parameters_path
    tag_names: list[str] = [SequencingFileTag.RUN_PARAMETERS, run_directory_data.id]
    hk_api.add_file_to_bundle_if_non_existent(
        file_path=run_parameters_file_path, bundle_name=run_directory_data.id, tag_names=tag_names
    )
    LOG.info(
        f"Added run parameters file {run_parameters_file_path} to {run_directory_data.id}"
        " in Housekeeper."
    )


def delete_sequencing_data_from_housekeeper(flow_cell_id: str, hk_api: HousekeeperAPI) -> None:
    """Delete FASTQ, SPRING and metadata files associated with a flow cell from Housekeeper."""
    tag_combinations: list[set[str]] = [
        {SequencingFileTag.FASTQ, flow_cell_id},
        {SequencingFileTag.SPRING, flow_cell_id},
        {SequencingFileTag.SPRING_METADATA, flow_cell_id},
    ]
    for tags in tag_combinations:
        housekeeper_files: Iterable[File] = hk_api.files(tags=tags)
        for housekeeper_file in housekeeper_files:
            hk_api.delete_file(file_id=housekeeper_file.id)



"""Module to test the PacBioStoreService."""

from unittest import mock

import pytest

from cg.services.run_devices.exc import (
    PostProcessingStoreDataError,
    PostProcessingDataTransferError,
)
from cg.services.run_devices.pacbio.data_storage_service.pacbio_store_service import (
    PacBioStoreService,
)
from cg.services.run_devices.pacbio.data_transfer_service.data_transfer_service import (
    PacBioDataTransferService,
)
from cg.services.run_devices.pacbio.data_transfer_service.dto import PacBioDTOs
from cg.services.run_devices.pacbio.run_data_generator.run_data import PacBioRunData
from cg.store.models import PacBioSampleSequencingMetrics, PacBioSequencingRun, PacBioSMRTCell

from cg.store.store import Store


def test_store_post_processing_data(
    pac_bio_store_service: PacBioStoreService,
    pac_bio_dtos: PacBioDTOs,
    expected_pac_bio_run_data: PacBioRunData,
):
    # GIVEN a PacBioStoreService

    # GIVEN a successful data transfer service

    # WHEN storing data for a PacBio instrument run
    with mock.patch(
        "cg.services.run_devices.pacbio.data_transfer_service.data_transfer_service.PacBioDataTransferService.get_post_processing_dtos",
        return_value=pac_bio_dtos,
    ):
        pac_bio_store_service.store_post_processing_data(expected_pac_bio_run_data)

    # THEN the SMRT cell data is stored
    smrt_cell: PacBioSMRTCell = pac_bio_store_service.store._get_query(PacBioSMRTCell).first()
    assert smrt_cell
    assert smrt_cell.internal_id == pac_bio_dtos.run_device.internal_id

    # THEN the sequencing run is stored
    sequencing_run: PacBioSequencingRun = pac_bio_store_service.store._get_query(
        PacBioSequencingRun
    ).first()
    assert sequencing_run
    assert sequencing_run.well == pac_bio_dtos.sequencing_run.well

    # THEN the sample sequencing metrics are stored
    sample_sequencing_run_metrics: PacBioSampleSequencingMetrics = (
        pac_bio_store_service.store._get_query(PacBioSampleSequencingMetrics).first()
    )
    assert sample_sequencing_run_metrics
    assert (
        sample_sequencing_run_metrics.sample.internal_id
        == pac_bio_dtos.sample_sequencing_metrics[0].sample_internal_id
    )


def test_store_post_processing_data_error_database(
    pac_bio_store_service: PacBioStoreService,
    pac_bio_dtos: PacBioDTOs,
    expected_pac_bio_run_data: PacBioRunData,
):
    # GIVEN a PacBioStoreService

    # GIVEN a successful data transfer service

    # WHEN storing data for a PacBio instrument run
    with mock.patch(
        "cg.services.run_devices.pacbio.data_transfer_service.data_transfer_service.PacBioDataTransferService.get_post_processing_dtos",
        return_value=pac_bio_dtos,
    ), mock.patch.object(Store, "create_pac_bio_smrt_cell", side_effect=ValueError):
        with pytest.raises(PostProcessingStoreDataError):
            pac_bio_store_service.store_post_processing_data(expected_pac_bio_run_data)


def test_store_post_processing_data_error_parser(
    pac_bio_store_service: PacBioStoreService,
    pac_bio_dtos: PacBioDTOs,
    expected_pac_bio_run_data: PacBioRunData,
):
    # GIVEN a PacBioStoreService

    # GIVEN a successful data transfer service

    # WHEN storing data for a PacBio instrument run
    with mock.patch(
        "cg.services.run_devices.pacbio.data_transfer_service.data_transfer_service.PacBioDataTransferService.get_post_processing_dtos",
        return_value=pac_bio_dtos,
    ), mock.patch.object(
        PacBioDataTransferService,
        "get_post_processing_dtos",
        side_effect=PostProcessingDataTransferError,
    ):
        with pytest.raises(PostProcessingStoreDataError):
            pac_bio_store_service.store_post_processing_data(expected_pac_bio_run_data)



from datetime import datetime
from unittest.mock import Mock

import pytest

from cg.constants.devices import DeviceType
from cg.services.run_devices.pacbio.data_storage_service.pacbio_store_service import (
    PacBioStoreService,
)
from cg.services.run_devices.pacbio.data_transfer_service.data_transfer_service import (
    PacBioDataTransferService,
)
from cg.services.run_devices.pacbio.data_transfer_service.dto import (
    PacBioDTOs,
    PacBioSampleSequencingMetricsDTO,
    PacBioSequencingRunDTO,
    PacBioSMRTCellDTO,
)
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


@pytest.fixture
def pac_bio_smrt_cell_dto() -> PacBioSMRTCellDTO:
    return PacBioSMRTCellDTO(type=DeviceType.PACBIO, internal_id="internal_id")


@pytest.fixture
def pac_bio_sequencing_run_dto() -> PacBioSequencingRunDTO:
    sample_data = {
        "type": DeviceType.PACBIO,
        "well": "A1",
        "plate": 1,
        "started_at": datetime.now(),
        "completed_at": datetime.now(),
        "movie_time_hours": 10,
        "hifi_reads": 500000,
        "hifi_yield": 3000000000,
        "hifi_mean_read_length": 6000,
        "hifi_median_read_length": 6000,
        "hifi_mean_length_n50": 5000,
        "hifi_median_read_quality": "Q20",
        "percent_reads_passing_q30": 99.5,
        "productive_zmws": 150000,
        "p0_percent": 0.5,
        "p1_percent": 1.5,
        "p2_percent": 98.0,
        "polymerase_mean_read_length": 7000,
        "polymerase_read_length_n50": 6500,
        "polymerase_mean_longest_subread": 12000,
        "polymerase_longest_subread_n50": 11000,
        "control_reads": 10000,
        "control_mean_read_length": 5000,
        "control_mean_read_concordance": 99.0,
        "control_mode_read_concordance": 99.5,
        "failed_reads": 1000,
        "failed_yield": 500000000,
        "failed_mean_read_length": 3000,
        "movie_name": "movie123",
    }
    return PacBioSequencingRunDTO(**sample_data)


@pytest.fixture
def pac_bio_sample_sequencing_metrics_dto() -> list[PacBioSampleSequencingMetricsDTO]:
    sample_metrics_data = {
        "sample_internal_id": "sample_123",
        "hifi_reads": 450000,
        "hifi_yield": 2750000000,
        "hifi_mean_read_length": 6100,
        "hifi_median_read_length": 6000,
        "hifi_median_read_quality": "Q30",
        "percent_reads_passing_q30": 98.6,
        "failed_reads": 1500,
        "failed_yield": 500000000,
        "failed_mean_read_length": 3200,
    }
    return [PacBioSampleSequencingMetricsDTO(**sample_metrics_data)]


@pytest.fixture
def pac_bio_dtos(
    pac_bio_smrt_cell_dto: PacBioSMRTCellDTO,
    pac_bio_sequencing_run_dto: PacBioSequencingRunDTO,
    pac_bio_sample_sequencing_metrics_dto: list[PacBioSampleSequencingMetricsDTO],
) -> PacBioDTOs:
    return PacBioDTOs(
        run_device=pac_bio_smrt_cell_dto,
        sequencing_run=pac_bio_sequencing_run_dto,
        sample_sequencing_metrics=pac_bio_sample_sequencing_metrics_dto,
    )


@pytest.fixture
def pac_bio_store_service(store: Store, helpers: StoreHelpers, pac_bio_dtos: PacBioDTOs):
    helpers.add_sample(
        store=store, internal_id=pac_bio_dtos.sample_sequencing_metrics[0].sample_internal_id
    )
    return PacBioStoreService(
        store=store, data_transfer_service=PacBioDataTransferService(metrics_service=Mock())
    )



import logging
from pathlib import Path

import pytest
from click.testing import CliRunner
from psutil import Process

from cg.cli.backup import backup_illumina_runs, encrypt_illumina_runs, fetch_illumina_run
from cg.constants import EXIT_SUCCESS, FileExtensions, SequencingRunDataAvailability
from cg.exc import IlluminaRunEncryptionError
from cg.models.cg_config import CGConfig
from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData
from cg.store.models import IlluminaSequencingRun
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


def test_backup_illumina_runs(
    cli_runner: CliRunner,
    backup_context: CGConfig,
    caplog,
    novaseq_x_flow_cell_id: str,
    novaseq_x_flow_cell_full_name: str,
    helpers: StoreHelpers,
):
    """Test backing up an Illumina run in dry run mode."""
    caplog.set_level(logging.DEBUG)

    # GIVEN a sequencing run without back-up
    sequencing_run: IlluminaSequencingRun = (
        backup_context.status_db.get_illumina_sequencing_run_by_device_internal_id(
            novaseq_x_flow_cell_id
        )
    )
    assert not sequencing_run.has_backup

    # GIVEN an encrypted run
    run_encrypt_dir = Path(backup_context.encryption.encryption_dir, novaseq_x_flow_cell_full_name)
    run_encrypt_dir.mkdir(parents=True, exist_ok=True)
    Path(run_encrypt_dir, novaseq_x_flow_cell_id).with_suffix(FileExtensions.COMPLETE).touch()

    # WHEN backing up runs in dry run mode
    result = cli_runner.invoke(backup_illumina_runs, ["--dry-run"], obj=backup_context)

    # THEN exits without any errors
    assert result.exit_code == EXIT_SUCCESS


def test_backup_illumina_runs_when_dsmc_is_running(
    cli_runner: CliRunner,
    context_with_illumina_data: CGConfig,
    caplog,
    mocker,
):
    """Test backing-up Illumina runs in dry run mode when Dsmc processing has started."""
    caplog.set_level(logging.ERROR)

    # GIVEN an ongoing Dsmc process
    mocker.patch.object(Process, "name", return_value="dsmc")

    # WHEN backing up Illumina runs in dry run mode
    result = cli_runner.invoke(backup_illumina_runs, ["--dry-run"], obj=context_with_illumina_data)

    # THEN exits without any errors
    assert result.exit_code == EXIT_SUCCESS

    # THEN communicate too many Dsmc processes are already running
    assert "Too many Dsmc processes are already running" in caplog.text


def test_backup_illumina_run_when_run_already_has_backup(
    cli_runner: CliRunner,
    context_with_illumina_data: CGConfig,
    caplog,
    store_with_illumina_sequencing_data: Store,
    novaseq_x_flow_cell_id: str,
    helpers: StoreHelpers,
):
    """Test backing-up an Illumina run in dry run mode when already backed-up."""
    caplog.set_level(logging.DEBUG)

    # GIVEN a store with a backed up sequencing run
    context_with_illumina_data.status_db_ = store_with_illumina_sequencing_data
    sequencing_run: IlluminaSequencingRun = (
        store_with_illumina_sequencing_data.get_illumina_sequencing_run_by_device_internal_id(
            novaseq_x_flow_cell_id
        )
    )
    sequencing_run.has_backup = True

    # WHEN backing up Illumina runs in dry run mode
    result = cli_runner.invoke(backup_illumina_runs, ["--dry-run"], obj=context_with_illumina_data)

    # THEN exits without any errors
    assert result.exit_code == EXIT_SUCCESS

    # THEN communicate the sequencing run has already benn backed upped
    assert (
        f"Sequencing run for flow cell: {novaseq_x_flow_cell_id} is already backed-up"
        in caplog.text
    )


def test_backup_illumina_runs_when_encryption_is_not_completed(
    cli_runner: CliRunner,
    context_with_illumina_data: CGConfig,
    caplog,
    novaseq_x_flow_cell_id: str,
):
    """Test backing-up Illumina run in dry run mode when encryption is not complete."""
    caplog.set_level(logging.DEBUG)

    # WHEN backing up Illumina runs in dry run mode
    result = cli_runner.invoke(backup_illumina_runs, ["--dry-run"], obj=context_with_illumina_data)

    # THEN exits without any errors
    assert result.exit_code == EXIT_SUCCESS

    # THEN communicate runs encryption is not completed
    assert (
        f"Sequencing run for flow cell: {novaseq_x_flow_cell_id} encryption process is not complete"
        in caplog.text
    )


def test_encrypt_illumina_runs(
    cli_runner: CliRunner, context_with_illumina_data: CGConfig, caplog, sbatch_job_number: str
):
    """Test encrypt Illumina runs in dry run mode."""
    caplog.set_level(logging.INFO)

    # WHEN encrypting run in dry run mode
    result = cli_runner.invoke(encrypt_illumina_runs, ["--dry-run"], obj=context_with_illumina_data)

    # THEN exits without any errors
    assert result.exit_code == EXIT_SUCCESS

    # THEN communicate encryption job is submitted
    assert f"Run encryption running as job {sbatch_job_number}" in caplog.text


def test_encrypt_illumina_runs_when_already_backed_up(
    cli_runner: CliRunner,
    context_with_illumina_data: CGConfig,
    caplog,
    novaseq_x_flow_cell_id: str,
    store_with_illumina_sequencing_data: Store,
    helpers: StoreHelpers,
):
    """Test encrypt Illumina run in dry run mode when there is already a back-up."""
    caplog.set_level(logging.DEBUG)

    # GIVEN a sequencing run with a back-up
    context_with_illumina_data.status_db_ = store_with_illumina_sequencing_data
    sequencing_run: IlluminaSequencingRun = (
        store_with_illumina_sequencing_data.get_illumina_sequencing_run_by_device_internal_id(
            novaseq_x_flow_cell_id
        )
    )
    sequencing_run.has_backup = True

    # GIVEN a sequencing runs directory

    # WHEN encrypting runs in dry run mode
    result = cli_runner.invoke(encrypt_illumina_runs, ["--dry-run"], obj=context_with_illumina_data)

    # THEN exits without any errors
    assert result.exit_code == EXIT_SUCCESS

    # THEN communicate run is already backed-up
    assert f"Run: {novaseq_x_flow_cell_id} is already backed-up" in caplog.text


def test_encrypt_illumina_run_when_sequencing_not_done(
    cli_runner: CliRunner,
    context_with_illumina_data: CGConfig,
    caplog,
    mocker,
    novaseq_x_flow_cell_id: str,
):
    """Test encrypt Illumina runs in dry run mode when sequencing is not done."""
    caplog.set_level(logging.DEBUG)

    # GIVEN Illumina runs that are being sequenced
    mocker.patch.object(IlluminaRunDirectoryData, "is_sequencing_run_ready")
    IlluminaRunDirectoryData.is_sequencing_run_ready.return_value = False

    # GIVEN a sequencing runs directory

    # WHEN encrypting runs in dry run mode
    result = cli_runner.invoke(encrypt_illumina_runs, ["--dry-run"], obj=context_with_illumina_data)

    # THEN exits without any errors
    assert result.exit_code == EXIT_SUCCESS

    # THEN communicate the run is not ready
    assert f"Run: {novaseq_x_flow_cell_id} is not ready" in caplog.text


def test_encrypt_illumina_run_when_encryption_already_started(
    cli_runner: CliRunner,
    encryption_context: CGConfig,
    caplog,
    pdc_archiving_dir: Path,
    novaseq_x_flow_cell_id: str,
    novaseq_x_flow_cell_full_name: str,
    store_with_illumina_sequencing_data: Store,
    tmp_path: Path,
    mocker,
):
    """Test encrypt Illumina runs in dry run mode when pending file exists"""
    caplog.set_level(logging.DEBUG)

    # GIVEN Illumina runs that are ready
    mocker.patch.object(IlluminaRunDirectoryData, "is_sequencing_run_ready")
    IlluminaRunDirectoryData.is_sequencing_run_ready.return_value = True

    # GIVEN a pending flag file
    Path(
        encryption_context.encryption.encryption_dir,
        novaseq_x_flow_cell_full_name,
        novaseq_x_flow_cell_id,
    ).with_suffix(FileExtensions.PENDING).touch()

    # WHEN encrypting Illumina runs in dry run mode
    result = cli_runner.invoke(encrypt_illumina_runs, ["--dry-run"], obj=encryption_context)

    # THEN exits without any errors
    assert result.exit_code == EXIT_SUCCESS

    # THEN communicate flow cell encryption already started
    assert f"Encryption already started for run: {novaseq_x_flow_cell_id}" in caplog.text


def test_encrypt_illumina_run_when_encryption_already_completed(
    cli_runner: CliRunner,
    encryption_context: CGConfig,
    novaseq_x_flow_cell_full_name: str,
    novaseq_x_flow_cell_id: str,
    caplog,
    pdc_archiving_dir: Path,
    mocker,
):
    """Test encrypt Illumina runs in dry run mode when completed file exists"""
    caplog.set_level(logging.DEBUG)

    # GIVEN Illumina runs that are ready
    mocker.patch.object(IlluminaRunDirectoryData, "is_sequencing_run_ready")
    IlluminaRunDirectoryData.is_sequencing_run_ready.return_value = True

    # GIVEN a complete flag file
    Path(
        encryption_context.encryption.encryption_dir,
        novaseq_x_flow_cell_full_name,
        novaseq_x_flow_cell_id,
    ).with_suffix(FileExtensions.COMPLETE).touch()

    # WHEN encrypting Illumina runs in dry run mode
    result = cli_runner.invoke(encrypt_illumina_runs, ["--dry-run"], obj=encryption_context)

    # THEN exits without any errors
    assert result.exit_code == EXIT_SUCCESS

    # THEN communicate flow cell encryption already completed
    assert f"Encryption already completed for run: {novaseq_x_flow_cell_id}" in caplog.text


def test_run_fetch_sequencing_run_dry_run_no_run_specified(
    cli_runner: CliRunner, backup_context: CGConfig, caplog
):
    """Test fetching flow cell when no Illumina runs with correct status."""
    caplog.set_level(logging.INFO)

    # GIVEN a context with a backup_api
    assert "backup_api" in backup_context.meta_apis

    # GIVEN that there are no Illumina runs set to "requested" in status_db
    assert not backup_context.status_db.get_illumina_sequencing_runs_by_data_availability(
        data_availability=[SequencingRunDataAvailability.REQUESTED]
    )

    # WHEN running the fetch_illumina_run command without specifying any flow cell in dry run mode
    result = cli_runner.invoke(fetch_illumina_run, ["--dry-run"], obj=backup_context)

    # THEN assert that it exits without any problems
    assert result.exit_code == EXIT_SUCCESS

    # THEN assert that it is communicated that no Illumina runs are requested
    assert "No sequencing run requested" in caplog.text


def test_run_fetch_sequencing_run_dry_run_retrieval_time(
    cli_runner: CliRunner, backup_context: CGConfig, caplog, mocker
):
    """Test fetching Illumina run retrieval time."""
    caplog.set_level(logging.INFO)

    # GIVEN a context with a backup_api
    assert "backup_api" in backup_context.meta_apis

    # GIVEN that there are no Illumina runs set to "requested" in status_db
    assert not backup_context.status_db.get_illumina_sequencing_runs_by_data_availability(
        data_availability=[SequencingRunDataAvailability.REQUESTED]
    )

    # GIVEN that the backup api returns a retrieval time
    expected_time = 60
    mocker.patch(
        "cg.services.illumina.backup.backup_service.IlluminaBackupService.fetch_sequencing_run",
        return_value=expected_time,
    )

    # WHEN running the fetch Illumina run command without specifying any illumina in dry run mode
    result = cli_runner.invoke(fetch_illumina_run, ["--dry-run"], obj=backup_context)

    # THEN assert that it exits without any problems
    assert result.exit_code == EXIT_SUCCESS

    # THEN assert that it is communicated that a retrieval time was found
    assert "Retrieval time" in caplog.text


def test_run_fetch_illumina_run_non_existing_flow_cell(
    cli_runner: CliRunner, backup_context: CGConfig, caplog
):
    # GIVEN a context with a backup api

    # GIVEN a non-existing flow cell id
    flow_cell_id = "non-existing-id"

    # WHEN running the command with the non-existing flow cell id
    result = cli_runner.invoke(
        fetch_illumina_run, ["--flow-cell-id", flow_cell_id], obj=backup_context
    )

    # THEN assert that it exits with a non-zero exit code
    assert result.exit_code != 0



"""
Module for deliver and rsync customer inbox on the HPC to customer inbox on the delivery server.
"""

import logging
import os
from pathlib import Path

from cg.constants.delivery import INBOX_NAME
from cg.exc import CgError
from cg.meta.meta import MetaAPI
from cg.models.cg_config import CGConfig
from cg.services.fastq_concatenation_service.fastq_concatenation_service import (
    FastqConcatenationService,
)
from cg.meta.deliver.fastq_path_generator import (
    generate_forward_concatenated_fastq_delivery_path,
    generate_reverse_concatenated_fastq_delivery_path,
)
from cg.store.models import Case

LOG = logging.getLogger(__name__)
PREFIX_TO_CONCATENATE = ["MWG", "MWL", "MWM", "MWR", "MWX"]


class DeliverTicketAPI(MetaAPI):
    def __init__(self, config: CGConfig):
        super().__init__(config)
        self.delivery_path: Path = Path(config.delivery_path)
        self.fastq_concatenation_service = FastqConcatenationService()

    def get_all_cases_from_ticket(self, ticket: str) -> list[Case]:
        return self.status_db.get_cases_by_ticket_id(ticket_id=ticket)

    def get_inbox_path(self, ticket: str) -> Path:
        cases: list[Case] = self.get_all_cases_from_ticket(ticket=ticket)
        if not cases:
            raise CgError(
                f"The customer id was not identified since no cases for ticket {ticket} was found"
            )
        customer_id: str = cases[0].customer.internal_id
        return Path(self.delivery_path, customer_id, INBOX_NAME, ticket)

    def check_if_upload_is_needed(self, ticket: str) -> bool:
        customer_inbox: Path = self.get_inbox_path(ticket=ticket)
        LOG.info(f"Checking if path exist: {customer_inbox}")
        if customer_inbox.exists():
            LOG.info(f"Could find path: {customer_inbox}")
            return False
        LOG.info(f"Could not find path: {customer_inbox}")
        return True

    def get_samples_from_ticket(self, ticket: str) -> list[str]:
        all_samples = []
        cases: list[Case] = self.get_all_cases_from_ticket(ticket=ticket)
        for case in cases:
            for link_obj in case.links:
                all_samples.append(link_obj.sample.name)
        return all_samples

    def report_missing_samples(self, ticket: str, dry_run: bool) -> None:
        customer_inbox: Path = self.get_inbox_path(ticket=ticket)
        missing_samples = []
        all_samples: list[str] = self.get_samples_from_ticket(ticket=ticket)
        if not customer_inbox.exists() and dry_run:
            LOG.info(f"Dry run, will not search for missing data in: {customer_inbox}")
            return
        if not customer_inbox.exists():
            LOG.info(
                f"The path {customer_inbox} do not exist, no search for missing data will be done"
            )
            return
        for dir_path in customer_inbox.iterdir():
            if len(os.listdir(dir_path)) == 0 and os.path.basename(dir_path) in all_samples:
                missing_samples.append(os.path.basename(dir_path))
        if len(missing_samples) > 0:
            LOG.info("No data delivered for sample(s):")
            for sample in missing_samples:
                LOG.info(sample)
        else:
            LOG.info("Data has been delivered for all samples")

    def concatenate_fastq_files(self, ticket: str, dry_run: bool) -> None:
        customer_inbox: Path = self.get_inbox_path(ticket=ticket)

        if not customer_inbox.exists() and dry_run:
            LOG.info(f"Dry run, nothing will be concatenated in: {customer_inbox}")
            return
        if not customer_inbox.exists():
            LOG.info(f"The path {customer_inbox} do not exist, nothing will be concatenated")
            return
        for dir_path in customer_inbox.iterdir():
            if len(os.listdir(dir_path)) == 0:
                LOG.info(f"Empty folder found: {dir_path}")
                continue
            if not dir_path.is_dir():
                continue
            forward_output_path = generate_forward_concatenated_fastq_delivery_path(
                fastq_directory=dir_path, sample_name=dir_path.name
            )
            reverse_output_path = generate_reverse_concatenated_fastq_delivery_path(
                fastq_directory=dir_path, sample_name=dir_path.name
            )
            self.fastq_concatenation_service.concatenate(
                fastq_directory=dir_path,
                forward_output_path=forward_output_path,
                reverse_output_path=reverse_output_path,
                remove_raw=True,
            )



from cg.services.orders.store_order_services.store_microbial_order import StoreMicrobialOrderService
from cg.services.orders.submitters.order_submitter import OrderSubmitter
from cg.services.orders.validate_order_services.validate_microbial_order import (
    ValidateMicrobialOrderService,
)


class MicrobialOrderSubmitter(OrderSubmitter):
    """
    Class for submitting microbial orders.
    This class is used to submit orders for the following workflows:
    - Sars-Cov-2
    - Microsalt
    - Mutant
    """

    def __init__(
        self,
        order_validation_service: ValidateMicrobialOrderService,
        order_store_service: StoreMicrobialOrderService,
    ):
        self.order_validation_service = order_validation_service
        self.order_store_service = order_store_service

    def submit_order(self, order_in) -> dict:
        """Submit a microbial order."""
        self.order_validation_service.validate_order(order_in)
        return self.order_store_service.store_order(order_in)



from pathlib import Path

import pytest

from cg.meta.orders import OrdersAPI
from cg.meta.orders.ticket_handler import TicketHandler
from cg.services.orders.submitters.order_submitter_registry import (
    OrderSubmitterRegistry,
    setup_order_submitter_registry,
)
from cg.store.store import Store
from tests.mocks.limsmock import MockLimsAPI
from tests.mocks.osticket import MockOsTicket


@pytest.fixture(scope="function")
def orders_api(
    base_store,
    osticket: MockOsTicket,
    lims_api: MockLimsAPI,
    order_submitter_registry: OrderSubmitterRegistry,
) -> OrdersAPI:
    return OrdersAPI(
        lims=lims_api,
        status=base_store,
        osticket=osticket,
        submitter_registry=order_submitter_registry,
    )


@pytest.fixture
def ticket_handler(store: Store, osticket: MockOsTicket) -> TicketHandler:
    return TicketHandler(status_db=store, osticket_api=osticket)


@pytest.fixture
def order_submitter_registry(base_store: Store, lims_api: MockLimsAPI) -> OrderSubmitterRegistry:
    return setup_order_submitter_registry(lims=lims_api, status_db=base_store)



from datetime import datetime
from pydantic import BaseModel

from cg.constants.subject import Sex


class ApplicationVersionDTO(BaseModel):
    id: int
    version: int | None = None
    valid_from: datetime | None = None
    price_standard: int | None = None
    price_priority: int | None = None
    price_express: int | None = None
    price_research: int | None = None
    price_clinical_trials: int | None = None
    comment: str | None = None
    created_at: datetime | None = None
    updated_at: datetime | None = None
    application_id: int | None = None


class ApplicationDTO(BaseModel):
    id: int
    tag: str | None = None
    prep_category: str | None = None
    is_external: bool | None = None
    description: str | None = None
    is_accredited: bool | None = None
    turnaround_time: int | None = None
    minimum_order: int | None = None
    sequencing_depth: int | None
    min_sequencing_depth: int | None = None
    target_reads: int | None = None
    percent_reads_guaranteed: int | None = None
    sample_amount: int | None = None
    sample_volume: str | None = None
    sample_concentration: str | None = None
    sample_concentration_minimum: float | None = None
    sample_concentration_maximum: float | None = None
    sample_concentration_minimum_cfdna: float | None = None
    sample_concentration_maximum_cfdna: float | None = None
    priority_processing: bool | None = None
    details: str | None = None
    limitations: str | None = None
    percent_kth: int | None = None
    comment: str | None = None
    is_archived: bool | None = None
    created_at: datetime | None = None
    updated_at: datetime | None = None


class CustomerDto(BaseModel):
    internal_id: str
    name: str


class SampleDTO(BaseModel):
    application: ApplicationDTO | None = None
    application_version: ApplicationVersionDTO | None = None
    name: str | None = None
    internal_id: str | None = None
    data_analysis: str | None = None
    data_delivery: str | None = None
    mother: str | None = None
    father: str | None = None
    family_name: str | None = None
    case_internal_id: str | None = None
    require_qc_ok: bool | None = None
    sex: Sex | None = None
    source: str | None = None
    priority: str | None = None
    formalin_fixation_time: int | None = None
    post_formalin_fixation_time: int | None = None
    tissue_block_size: str | None = None
    cohorts: list[str] | None = None
    phenotype_groups: list[str] | None = None
    phenotype_terms: list[str] | None = None
    subject_id: str | None = None
    synopsis: str | None = None
    age_at_sampling: int | None = None
    comment: str | None = None
    control: str | None = None
    elution_buffer: str | None = None
    container: str | None = None
    container_name: str | None = None
    well_position: str | None = None
    volume: int | None = None
    concentration_ng_ul: int | None = None
    panels: list[str] | None = None
    status: str | None = None
    tumour: bool | None = None
    reference_genome: str | None = None
    customer: CustomerDto | None = None


class SamplesResponse(BaseModel):
    samples: list[SampleDTO]
    total: int



"""Module to handle dictionary helper functions."""


def remove_duplicate_dicts(dicts: list[dict]) -> list[dict]:
    return [
        dict(dictionary_tuple)
        for dictionary_tuple in {tuple(dictionary.items()) for dictionary in dicts}
    ]


def get_list_from_dictionary(dictionary: dict) -> list:
    """Return a list of the passed dict non-empty key values."""
    list_from_dict: list[str] = []
    for key, value in dictionary.items():
        if value:
            list_from_dict.extend([key, value])
    return list_from_dict


def get_full_path_dictionary(dictionary: dict) -> dict:
    """Return a full path of the values of the passed dictionary."""
    return {key: value.full_path if value else None for key, value in dictionary.items()}



import pytest

from cg.models.cg_config import CGConfig, PostProcessingServices


@pytest.fixture
def pac_bio_context(
    cg_context: CGConfig, post_processing_services: PostProcessingServices
) -> CGConfig:
    cg_context.post_processing_services_ = post_processing_services
    return cg_context



import logging

import pytest
from _pytest.fixtures import FixtureRequest
from _pytest.logging import LogCaptureFixture
from click.testing import CliRunner

from cg.apps.lims import LimsAPI
from cg.cli.workflow.base import workflow as workflow_cli
from cg.constants import EXIT_SUCCESS, Workflow
from cg.constants.nextflow import NEXTFLOW_WORKFLOWS
from cg.meta.workflow.raredisease import RarediseaseAnalysisAPI
from cg.meta.workflow.nf_analysis import NfAnalysisAPI
from cg.models.cg_config import CGConfig
from tests.cli.workflow.conftest import mock_analysis_illumina_run


@pytest.mark.parametrize(
    "workflow",
    NEXTFLOW_WORKFLOWS,
)
def test_start(
    cli_runner: CliRunner,
    workflow: Workflow,
    caplog: LogCaptureFixture,
    mock_analysis_illumina_run,
    request: FixtureRequest,
    mocker,
):
    """Test to ensure all parts of start command will run successfully given ideal conditions."""
    caplog.set_level(logging.DEBUG)
    context: CGConfig = request.getfixturevalue(f"{workflow}_context")

    # GIVEN a case id
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")

    # GIVEN a mocked config

    # GIVEN decompression is not needed
    mocker.patch.object(NfAnalysisAPI, "resolve_decompression", return_value=None)

    # GIVEN that the sample source in LIMS is set
    mocker.patch.object(LimsAPI, "get_source", return_value="blood")

    # WHEN invoking the command with dry-run specified
    result = cli_runner.invoke(workflow_cli, [workflow, "start", case_id, "--dry-run"], obj=context)

    # THEN command should execute successfully
    assert result.exit_code == EXIT_SUCCESS
    assert case_id in caplog.text

    # THEN command should not include resume flag
    assert "-resume" not in caplog.text


@pytest.mark.parametrize(
    "workflow",
    NEXTFLOW_WORKFLOWS,
)
def test_start_available(
    cli_runner: CliRunner,
    workflow: Workflow,
    caplog: LogCaptureFixture,
    mocker,
    mock_analysis_illumina_run,
    request: FixtureRequest,
    case_id_not_enough_reads: str,
):
    """
    Test to ensure all parts of compound start-available command are executed given ideal conditions.
    Test that start-available picks up eligible cases and does not pick up ineligible ones.
    """
    caplog.set_level(logging.DEBUG)
    context: CGConfig = request.getfixturevalue(f"{workflow}_context")

    # GIVEN a case with enough reads and another case with insufficient reads
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")

    # GIVEN a mocked config

    # GIVEN decompression is not needed
    mocker.patch.object(NfAnalysisAPI, "resolve_decompression", return_value=None)

    # GIVEN that the sample source in LIMS is set
    mocker.patch.object(LimsAPI, "get_source", return_value="blood")

    # WHEN invoking the command with dry-run specified
    result = cli_runner.invoke(
        workflow_cli, [workflow, "start-available", "--dry-run"], obj=context
    )

    # THEN command exits successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN it should successfully identify the one case eligible for auto-start
    assert case_id in caplog.text

    # THEN the case without enough reads should not start
    assert case_id_not_enough_reads not in caplog.text



"""All models aggregated in a base class."""

from dataclasses import dataclass
from typing import Type

from sqlalchemy import and_, func
from sqlalchemy.orm import Query, Session

from cg.store.models import Analysis, Application, ApplicationLimitations, ApplicationVersion
from cg.store.models import Base as ModelBase
from cg.store.models import (
    Case,
    CaseSample,
    Customer,
    IlluminaFlowCell,
    IlluminaSampleSequencingMetrics,
    IlluminaSequencingRun,
    Sample,
)


@dataclass
class BaseHandler:
    """All queries in one base class."""

    def __init__(self, session: Session):
        self.session = session

    def _get_query(self, table: Type[ModelBase]) -> Query:
        """Return a query for the given table."""
        return self.session.query(table)

    def _get_outer_join_cases_with_analyses_query(self) -> Query:
        """Return a query for all cases in the database with an analysis."""
        return (
            self._get_query(table=Case)
            .outerjoin(Analysis)
            .join(Case.links)
            .join(CaseSample.sample)
            .join(ApplicationVersion)
            .join(Application)
        )

    def _get_join_cases_with_samples_query(self) -> Query:
        """Return a join query for all cases in the database with samples."""
        return (
            self._get_query(table=Case).join(Case.links).join(CaseSample.sample).join(Case.customer)
        )

    def _get_join_analysis_case_query(self) -> Query:
        """Return join analysis case query."""
        return self._get_query(table=Analysis).join(Analysis.case)

    def _get_join_case_sample_query(self) -> Query:
        """Return join case sample query."""
        return self._get_query(table=CaseSample).join(CaseSample.case).join(CaseSample.sample)

    def _join_sample_and_case(self):
        return (
            self._get_query(Case)
            .join(CaseSample, Case.id == CaseSample.case_id)
            .join(Sample, CaseSample.sample_id == Sample.id)
        )

    def _get_join_case_and_sample_query(self) -> Query:
        """Return join case sample query."""
        return self._get_query(table=Case).join(Case.links).join(CaseSample.sample)

    def _get_join_sample_and_customer_query(self) -> Query:
        """Return join sample and customer query."""
        return self._get_query(table=Sample).join(Customer)

    def _get_join_sample_family_query(self) -> Query:
        """Return a join sample case relationship query."""
        return self._get_query(table=Sample).join(Case.links).join(CaseSample.sample)

    def _get_join_sample_application_version_query(self) -> Query:
        """Return join sample to application version query."""
        return (
            self._get_query(table=Sample)
            .join(Sample.application_version)
            .join(ApplicationVersion.application)
        )

    def _get_join_analysis_sample_family_query(self) -> Query:
        """Return join analysis to sample to case query."""
        return self._get_query(table=Analysis).join(Case).join(Case.links).join(CaseSample.sample)

    def _get_subquery_with_latest_case_analysis_date(self) -> Query:
        """Return a subquery with the case internal id and the date of its latest analysis."""
        case_and_date: Query = (
            self._get_join_analysis_case_query()
            .group_by(Case.id)
            .with_entities(Analysis.case_id, func.max(Analysis.started_at).label("started_at"))
            .subquery()
        )
        return case_and_date

    def _get_latest_analyses_for_cases_query(self) -> Query:
        """Return a join query for the latest analysis for each case."""
        analyses: Query = self._get_query(table=Analysis)
        case_and_date_subquery: Query = self._get_subquery_with_latest_case_analysis_date()
        return analyses.join(
            case_and_date_subquery,
            and_(
                Analysis.case_id == case_and_date_subquery.c.case_id,
                Analysis.started_at == case_and_date_subquery.c.started_at,
            ),
        )

    def _get_join_application_limitations_query(self) -> Query:
        """Return a join query for all application limitations."""
        return self._get_query(table=ApplicationLimitations).join(
            ApplicationLimitations.application
        )

    def _get_joined_illumina_sample_tables(self) -> Query:
        """Return a join query for all Illumina tables and Sample."""
        return (
            self._get_query(table=IlluminaSampleSequencingMetrics)
            .join(Sample)
            .join(IlluminaSequencingRun)
            .join(IlluminaFlowCell)
        )



from http import HTTPStatus
from typing import Any

from flask import Blueprint, abort, jsonify, make_response

from cg.server.endpoints.utils import before_request, is_public
from cg.server.ext import db
from cg.store.models import (
    Application,
    ApplicationLimitations,
)

APPLICATIONS_BLUEPRINT = Blueprint("applications", __name__, url_prefix="/api/v1")
APPLICATIONS_BLUEPRINT.before_request(before_request)


@APPLICATIONS_BLUEPRINT.route("/applications")
@is_public
def get_applications():
    """Return application tags."""
    applications: list[Application] = db.get_applications_is_not_archived()
    parsed_applications: list[dict] = [application.to_dict() for application in applications]
    return jsonify(applications=parsed_applications)


@APPLICATIONS_BLUEPRINT.route("/applications/<tag>")
@is_public
def get_application(tag: str):
    """Return an application tag."""
    application: Application = db.get_application_by_tag(tag=tag)
    if not application:
        return abort(make_response(jsonify(message="Application not found"), HTTPStatus.NOT_FOUND))

    application_limitations: list[ApplicationLimitations] = db.get_application_limitations_by_tag(
        tag
    )
    application_dict: dict[str, Any] = application.to_dict()
    application_dict["workflow_limitations"] = [
        limitation.to_dict() for limitation in application_limitations
    ]
    return jsonify(**application_dict)


@APPLICATIONS_BLUEPRINT.route("/applications/<tag>/workflow_limitations")
@is_public
def get_application_workflow_limitations(tag: str):
    """Return application workflow specific limitations."""
    if application_limitations := db.get_application_limitations_by_tag(tag):
        return jsonify([limitation.to_dict() for limitation in application_limitations])
    else:
        return jsonify(message="Application limitations not found"), HTTPStatus.NOT_FOUND



"""Utils related to string manipulation."""

from cg.exc import CgError


def get_element_from_split(value: str, element_position: int, split: str) -> str:
    elements: list[str] = value.split(split)
    if len(elements) < element_position:
        raise CgError(message="Provided element position out of bounds.")
    return elements[element_position]



import re
from datetime import datetime
from typing import Any, TypeVar

from pydantic import BaseModel, Field, field_validator, model_validator

from cg.constants.pacbio import (
    CCSAttributeIDs,
    ControlAttributeIDs,
    LoadingAttributesIDs,
    PolymeraseDataAttributeIDs,
    SmrtLinkDatabasesIDs,
)
from cg.services.run_devices.abstract_models import RunMetrics
from cg.utils.calculations import fraction_to_percent

BaseMetrics = TypeVar("BaseMetrics", bound=BaseModel)


class ReadMetrics(BaseModel):
    """Model for the read metrics."""

    hifi_reads: int = Field(..., alias=CCSAttributeIDs.HIFI_READS)
    hifi_yield: int = Field(..., alias=CCSAttributeIDs.HIFI_YIELD)
    hifi_mean_read_length: int = Field(..., alias=CCSAttributeIDs.HIFI_MEAN_READ_LENGTH)
    hifi_median_read_length: int = Field(..., alias=CCSAttributeIDs.HIFI_MEDIAN_READ_LENGTH)
    hifi_mean_length_n50: int = Field(..., alias=CCSAttributeIDs.HIFI_READ_LENGTH_N50)
    hifi_median_read_quality: int = Field(..., alias=CCSAttributeIDs.HIFI_MEDIAN_READ_QUALITY)
    percent_q30: float = Field(..., alias=CCSAttributeIDs.PERCENT_Q30)
    failed_reads: int = Field(..., alias=CCSAttributeIDs.FAILED_READS)
    failed_yield: int = Field(..., alias=CCSAttributeIDs.FAILED_YIELD)
    failed_mean_read_length: int = Field(..., alias=CCSAttributeIDs.FAILED_MEAN_READ_LENGTH)


class ControlMetrics(BaseModel):
    """Model for the control metrics."""

    reads: int = Field(..., alias=ControlAttributeIDs.NUMBER_OF_READS)
    mean_read_length: int = Field(..., alias=ControlAttributeIDs.MEAN_READ_LENGTH)
    percent_mean_concordance_reads: float = Field(
        ..., alias=ControlAttributeIDs.PERCENT_MEAN_READ_CONCORDANCE
    )
    percent_mode_concordance_reads: float = Field(
        ..., alias=ControlAttributeIDs.PERCENT_MODE_READ_CONCORDANCE
    )

    _validate_percent_mean_concordance_reads = field_validator(
        "percent_mean_concordance_reads", mode="before"
    )(fraction_to_percent)

    _validate_percent_mode_concordance_reads = field_validator(
        "percent_mode_concordance_reads", mode="before"
    )(fraction_to_percent)


class ProductivityMetrics(BaseModel):
    """Model for the loading metrics."""

    productive_zmws: int = Field(..., alias=LoadingAttributesIDs.PRODUCTIVE_ZMWS)
    p_0: int = Field(..., alias=LoadingAttributesIDs.P_0)
    p_1: int = Field(..., alias=LoadingAttributesIDs.P_1)
    p_2: int = Field(..., alias=LoadingAttributesIDs.P_2)
    percent_p_0: float
    percent_p_1: float
    percent_p_2: float

    @model_validator(mode="before")
    @classmethod
    def set_percentages(cls, data: Any):
        if isinstance(data, dict):
            productive_zmws = data.get(LoadingAttributesIDs.PRODUCTIVE_ZMWS)
            p_0 = data.get(LoadingAttributesIDs.P_0)
            p_1 = data.get(LoadingAttributesIDs.P_1)
            p_2 = data.get(LoadingAttributesIDs.P_2)
            data["percent_p_0"] = round((p_0 / productive_zmws) * 100, 0)
            data["percent_p_1"] = round((p_1 / productive_zmws) * 100, 0)
            data["percent_p_2"] = round((p_2 / productive_zmws) * 100, 0)
        return data


class PolymeraseMetrics(BaseModel):
    """Model for the polymerase metrics."""

    mean_read_length: int = Field(..., alias=PolymeraseDataAttributeIDs.MEAN_READ_LENGTH)
    read_length_n50: int = Field(..., alias=PolymeraseDataAttributeIDs.READ_LENGTH_N50)
    mean_longest_subread_length: int = Field(
        ..., alias=PolymeraseDataAttributeIDs.MEAN_LONGEST_SUBREAD_LENGTH
    )
    longest_subread_length_n50: int = Field(
        ..., alias=PolymeraseDataAttributeIDs.LONGEST_SUBREAD_LENGTH_N50
    )


class SmrtlinkDatasetsMetrics(BaseModel):
    """Model to parse metrics in the SMRTlink datasets report."""

    cell_id: str = Field(..., alias=SmrtLinkDatabasesIDs.CELL_ID)
    well: str = Field(..., alias=SmrtLinkDatabasesIDs.WELL_NAME)
    well_sample_name: str = Field(..., alias=SmrtLinkDatabasesIDs.WELL_SAMPLE_NAME)
    sample_internal_id: str = Field(..., alias=SmrtLinkDatabasesIDs.BIO_SAMPLE_NAME)
    run_started_at: datetime
    run_completed_at: datetime = Field(..., alias=SmrtLinkDatabasesIDs.RUN_COMPLETED_AT)
    movie_name: str = Field(..., alias=SmrtLinkDatabasesIDs.MOVIE_NAME)
    cell_index: int = Field(..., alias=SmrtLinkDatabasesIDs.CELL_INDEX)
    path: str = Field(..., alias=SmrtLinkDatabasesIDs.PATH)
    plate: int

    @model_validator(mode="before")
    @classmethod
    def extract_plate(cls, data: Any):
        if isinstance(data, dict):
            path = data.get("path")
            if path:
                pattern = r"/([12])_[ABCD]01"
                match = re.search(pattern, path)
                if match:
                    data["plate"] = match.group(1)
        return data

    @model_validator(mode="before")
    @classmethod
    def set_sequencing_completed_at(cls, data: Any):
        if isinstance(data, dict):
            movie_name = data.get(SmrtLinkDatabasesIDs.MOVIE_NAME)
            if movie_name:
                date: str = movie_name.split("_")[1] + movie_name.split("_")[2]
                data["run_started_at"] = datetime.strptime(date, "%y%m%d%H%M%S")
        return data


class PacBioMetrics(RunMetrics):
    """Model that holds all relevant PacBio metrics."""

    read: ReadMetrics
    control: ControlMetrics
    productivity: ProductivityMetrics
    polymerase: PolymeraseMetrics
    dataset_metrics: SmrtlinkDatasetsMetrics



"""Fixtures for the PacBio post processing services."""

from pathlib import Path

import pytest

from cg.services.run_devices.pacbio.run_data_generator.run_data import PacBioRunData


@pytest.fixture
def expected_pac_bio_run_data(
    pac_bio_test_run_name: str, pac_bio_smrt_cell_dir_1_a01: Path
) -> PacBioRunData:
    return PacBioRunData(
        full_path=pac_bio_smrt_cell_dir_1_a01,
        sequencing_run_name=pac_bio_test_run_name,
        well_name="A01",
        plate=1,
    )



import logging
from pathlib import Path

from housekeeper.store.models import File

from cg.meta.clean.clean_retrieved_spring_files import CleanRetrievedSpringFilesAPI


def test_get_files_to_remove(
    populated_clean_retrieved_spring_files_api: CleanRetrievedSpringFilesAPI,
    path_to_old_retrieved_spring_file: str,
    path_to_old_spring_file_in_housekeeper: str,
):
    """Tests that only old retrieved files are cleaned with clean_retrieved_spring_files. With the provided populated
    api, this should not return a newly retrieved spring file, a fastq file nor an archived spring file which
    has not been retrieved."""

    # GIVEN a CleanRetrievedSpringFilesAPI with a populated Housekeeper database

    # WHEN getting files to remove when cleaning retrieved spring files
    files_to_remove: list[File] = populated_clean_retrieved_spring_files_api._get_files_to_remove(
        age_limit=7
    )

    # THEN only the file with an old enough 'retrieved_at' should be returned
    assert [file.path for file in files_to_remove] == [path_to_old_spring_file_in_housekeeper]


def test_clean_retrieved_spring_files_dry_run(
    populated_clean_retrieved_spring_files_api: CleanRetrievedSpringFilesAPI,
    path_to_old_retrieved_spring_file: str,
    retrieved_test_bundle_name: str,
    caplog,
    path_to_old_spring_file_in_housekeeper: str,
):
    """Tests that only the Spring file with an old enough 'retrieved_at' would be removed when cleaning retrieved
    Spring files."""

    caplog.set_level(logging.INFO)

    # GIVEN a CleanRetrievedSpringFilesAPI with a populated Housekeeper database
    # GIVEN that an old retrieved file exists
    files = populated_clean_retrieved_spring_files_api.housekeeper_api.files(
        bundle=retrieved_test_bundle_name
    ).all()
    for file in files:
        assert Path(file.full_path).exists()

    # WHEN running 'clean_retrieved_spring_files'
    populated_clean_retrieved_spring_files_api.clean_retrieved_spring_files(age_limit=7)

    # THEN only the file with an old enough 'retrieved_at' should have been removed
    for file in files:
        if file.path == path_to_old_spring_file_in_housekeeper:
            assert not Path(file.full_path).exists()
        else:
            assert Path(file.full_path).exists()



"""Sbatch templates for the rsync function"""

RSYNC_COMMAND = """
rsync -rvL {source_path} {destination_path}
"""

RSYNC_CONTENTS_COMMAND = """
rsync -rvL {source_path}/ {destination_path}
"""

COVID_RSYNC = """
rsync -rvtL {source_path} {destination_path}
rsync -rvtL --chmod=777 {covid_report_path} {covid_destination_path}
"""

ERROR_RSYNC_FUNCTION = """
echo "Rsync failed"
"""



from http import HTTPStatus

import pytest
from pytest_mock import MockFixture

from cg.clients.arnold.api import ArnoldAPIClient
from cg.clients.arnold.dto.create_case_request import CreateCaseRequest


@pytest.fixture
def arnold_config() -> dict:
    return {"arnold": {"api_url": "arnold_url"}}


@pytest.fixture
def arnold_client(arnold_config: dict) -> ArnoldAPIClient:
    return ArnoldAPIClient(arnold_config)


@pytest.fixture
def mock_post_request_ok(mocker: MockFixture) -> MockFixture:
    mocked_response = mocker.Mock()
    mocked_response.status_code = HTTPStatus.OK
    return mocked_response


@pytest.fixture
def error_content() -> str:
    return "Test content."


@pytest.fixture
def mock_post_request_not_found(mocker: MockFixture, error_content: str) -> MockFixture:
    mocked_response = mocker.Mock()
    mocked_response.status_code = HTTPStatus.NOT_FOUND
    mocked_response.content = error_content
    return mocked_response


@pytest.fixture
def create_case_request() -> CreateCaseRequest:
    return CreateCaseRequest(case_id="test_case", case_info={})



"""Module for PacBio fixtures returning Path objects."""

from pathlib import Path

import pytest

from cg.constants.pacbio import PacBioDirsAndFiles


# Directory fixtures
@pytest.fixture
def pac_bio_fixtures_dir(devices_dir: Path) -> Path:
    """Return the path to the PacBio fixtures directory."""
    return Path(devices_dir, "pacbio")


@pytest.fixture
def pac_bio_wrong_metrics_dir(pac_bio_fixtures_dir: Path) -> Path:
    """Return the path to the PacBio metrics directory."""
    return Path(pac_bio_fixtures_dir, "wrong_metrics")


@pytest.fixture
def pac_bio_runs_dir(pac_bio_fixtures_dir: Path) -> Path:
    """Return the path to the PacBio run directory."""
    return Path(pac_bio_fixtures_dir, "SMRTcells")


@pytest.fixture
def pac_bio_test_run_dir(pac_bio_runs_dir: Path, pac_bio_test_run_name: str) -> Path:
    """Return the path to a PacBio run directory."""
    return Path(pac_bio_runs_dir, pac_bio_test_run_name)


@pytest.fixture
def pac_bio_smrt_cell_dir_1_a01(pac_bio_test_run_dir: Path, pac_bio_smrt_cell_name: str) -> Path:
    """Return the path to a PacBio SMRT cell directory."""
    return Path(pac_bio_test_run_dir, pac_bio_smrt_cell_name)


@pytest.fixture
def pac_bio_hifi_reads_dir(pac_bio_smrt_cell_dir_1_a01: Path) -> Path:
    """Return the path to a PacBio HiFi reads directory."""
    return Path(pac_bio_smrt_cell_dir_1_a01, PacBioDirsAndFiles.HIFI_READS)


@pytest.fixture
def pac_bio_run_statistics_dir(pac_bio_smrt_cell_dir_1_a01: Path) -> Path:
    """Return the path to the PacBio SMRT cell statistics directory."""
    return Path(pac_bio_smrt_cell_dir_1_a01, PacBioDirsAndFiles.STATISTICS_DIR)


@pytest.fixture
def pac_bio_run_reports_dir(pac_bio_run_statistics_dir: Path) -> Path:
    """Return the path to the PacBio SMRT cell unzipped_reports directory"""
    return Path(pac_bio_run_statistics_dir, PacBioDirsAndFiles.UNZIPPED_REPORTS_DIR)


@pytest.fixture
def pac_bio_wrong_metrics_file(pac_bio_wrong_metrics_dir: Path) -> Path:
    """Return the path to a temporary PacBio statistics directory."""
    return Path(pac_bio_wrong_metrics_dir, "metrics.json")


# File fixtures


@pytest.fixture
def pac_bio_ccs_report_file(
    pac_bio_run_statistics_dir: Path, pac_bio_1_a01_cell_full_name: str
) -> Path:
    """Return the path to the PacBio CCS report file."""
    return Path(
        pac_bio_run_statistics_dir,
        f"{pac_bio_1_a01_cell_full_name}.{PacBioDirsAndFiles.CCS_REPORT_SUFFIX}",
    )


@pytest.fixture
def pac_bio_control_report_file(pac_bio_run_reports_dir: Path) -> Path:
    """Return the path to the PacBio control report file."""
    return Path(pac_bio_run_reports_dir, PacBioDirsAndFiles.CONTROL_REPORT)


@pytest.fixture
def pac_bio_loading_report_file(pac_bio_run_reports_dir: Path) -> Path:
    """Return the path to the PacBio loading report file."""
    return Path(pac_bio_run_reports_dir, PacBioDirsAndFiles.LOADING_REPORT)


@pytest.fixture
def pac_bio_raw_data_report_file(pac_bio_run_reports_dir: Path) -> Path:
    """Return the path to the PacBio raw data report file."""
    return Path(pac_bio_run_reports_dir, PacBioDirsAndFiles.RAW_DATA_REPORT)


@pytest.fixture
def pac_bio_smrtlink_datasets_report_file(pac_bio_run_reports_dir: Path) -> Path:
    """Return the path to the PacBio SMRTLink datasets report file."""
    return Path(pac_bio_run_reports_dir, PacBioDirsAndFiles.SMRTLINK_DATASETS_REPORT)


@pytest.fixture
def pac_bio_report_files_to_parse(
    pac_bio_ccs_report_file: Path,
    pac_bio_control_report_file: Path,
    pac_bio_loading_report_file: Path,
    pac_bio_raw_data_report_file: Path,
    pac_bio_smrtlink_datasets_report_file: Path,
) -> list[Path]:
    """Return the list of PacBio report files to parse."""
    return [
        pac_bio_control_report_file,
        pac_bio_loading_report_file,
        pac_bio_raw_data_report_file,
        pac_bio_smrtlink_datasets_report_file,
        pac_bio_ccs_report_file,
    ]


@pytest.fixture
def pac_bio_hifi_read_file(pac_bio_hifi_reads_dir: Path, pac_bio_1_a01_cell_full_name: str) -> Path:
    """Return the PacBio HiFi read file."""
    return Path(pac_bio_hifi_reads_dir, f"{pac_bio_1_a01_cell_full_name}.hifi_reads.bam")
